from . import common_pb2 as _common_pb2
from . import schema_pb2 as _schema_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class InsertDataVersion(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RowBased: _ClassVar[InsertDataVersion]
    ColumnBased: _ClassVar[InsertDataVersion]
RowBased: InsertDataVersion
ColumnBased: InsertDataVersion

class InsertRequest(_message.Message):
    __slots__ = ("base", "shardName", "db_name", "collection_name", "partition_name", "dbID", "collectionID", "partitionID", "segmentID", "timestamps", "rowIDs", "row_data", "fields_data", "num_rows", "version", "namespace")
    BASE_FIELD_NUMBER: _ClassVar[int]
    SHARDNAME_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_NAME_FIELD_NUMBER: _ClassVar[int]
    PARTITION_NAME_FIELD_NUMBER: _ClassVar[int]
    DBID_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    PARTITIONID_FIELD_NUMBER: _ClassVar[int]
    SEGMENTID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
    ROWIDS_FIELD_NUMBER: _ClassVar[int]
    ROW_DATA_FIELD_NUMBER: _ClassVar[int]
    FIELDS_DATA_FIELD_NUMBER: _ClassVar[int]
    NUM_ROWS_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    shardName: str
    db_name: str
    collection_name: str
    partition_name: str
    dbID: int
    collectionID: int
    partitionID: int
    segmentID: int
    timestamps: _containers.RepeatedScalarFieldContainer[int]
    rowIDs: _containers.RepeatedScalarFieldContainer[int]
    row_data: _containers.RepeatedCompositeFieldContainer[_common_pb2.Blob]
    fields_data: _containers.RepeatedCompositeFieldContainer[_schema_pb2.FieldData]
    num_rows: int
    version: InsertDataVersion
    namespace: str
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., shardName: _Optional[str] = ..., db_name: _Optional[str] = ..., collection_name: _Optional[str] = ..., partition_name: _Optional[str] = ..., dbID: _Optional[int] = ..., collectionID: _Optional[int] = ..., partitionID: _Optional[int] = ..., segmentID: _Optional[int] = ..., timestamps: _Optional[_Iterable[int]] = ..., rowIDs: _Optional[_Iterable[int]] = ..., row_data: _Optional[_Iterable[_Union[_common_pb2.Blob, _Mapping]]] = ..., fields_data: _Optional[_Iterable[_Union[_schema_pb2.FieldData, _Mapping]]] = ..., num_rows: _Optional[int] = ..., version: _Optional[_Union[InsertDataVersion, str]] = ..., namespace: _Optional[str] = ...) -> None: ...

class DeleteRequest(_message.Message):
    __slots__ = ("base", "shardName", "db_name", "collection_name", "partition_name", "dbID", "collectionID", "partitionID", "int64_primary_keys", "timestamps", "num_rows", "primary_keys", "segment_id")
    BASE_FIELD_NUMBER: _ClassVar[int]
    SHARDNAME_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_NAME_FIELD_NUMBER: _ClassVar[int]
    PARTITION_NAME_FIELD_NUMBER: _ClassVar[int]
    DBID_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    PARTITIONID_FIELD_NUMBER: _ClassVar[int]
    INT64_PRIMARY_KEYS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
    NUM_ROWS_FIELD_NUMBER: _ClassVar[int]
    PRIMARY_KEYS_FIELD_NUMBER: _ClassVar[int]
    SEGMENT_ID_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    shardName: str
    db_name: str
    collection_name: str
    partition_name: str
    dbID: int
    collectionID: int
    partitionID: int
    int64_primary_keys: _containers.RepeatedScalarFieldContainer[int]
    timestamps: _containers.RepeatedScalarFieldContainer[int]
    num_rows: int
    primary_keys: _schema_pb2.IDs
    segment_id: int
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., shardName: _Optional[str] = ..., db_name: _Optional[str] = ..., collection_name: _Optional[str] = ..., partition_name: _Optional[str] = ..., dbID: _Optional[int] = ..., collectionID: _Optional[int] = ..., partitionID: _Optional[int] = ..., int64_primary_keys: _Optional[_Iterable[int]] = ..., timestamps: _Optional[_Iterable[int]] = ..., num_rows: _Optional[int] = ..., primary_keys: _Optional[_Union[_schema_pb2.IDs, _Mapping]] = ..., segment_id: _Optional[int] = ...) -> None: ...

class MsgPosition(_message.Message):
    __slots__ = ("channel_name", "msgID", "msgGroup", "timestamp")
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    MSGID_FIELD_NUMBER: _ClassVar[int]
    MSGGROUP_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    channel_name: str
    msgID: bytes
    msgGroup: str
    timestamp: int
    def __init__(self, channel_name: _Optional[str] = ..., msgID: _Optional[bytes] = ..., msgGroup: _Optional[str] = ..., timestamp: _Optional[int] = ...) -> None: ...

class CreateCollectionRequest(_message.Message):
    __slots__ = ("base", "db_name", "collectionName", "partitionName", "dbID", "collectionID", "partitionID", "schema", "virtualChannelNames", "physicalChannelNames", "partitionIDs")
    BASE_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONNAME_FIELD_NUMBER: _ClassVar[int]
    PARTITIONNAME_FIELD_NUMBER: _ClassVar[int]
    DBID_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    PARTITIONID_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    VIRTUALCHANNELNAMES_FIELD_NUMBER: _ClassVar[int]
    PHYSICALCHANNELNAMES_FIELD_NUMBER: _ClassVar[int]
    PARTITIONIDS_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    db_name: str
    collectionName: str
    partitionName: str
    dbID: int
    collectionID: int
    partitionID: int
    schema: bytes
    virtualChannelNames: _containers.RepeatedScalarFieldContainer[str]
    physicalChannelNames: _containers.RepeatedScalarFieldContainer[str]
    partitionIDs: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., db_name: _Optional[str] = ..., collectionName: _Optional[str] = ..., partitionName: _Optional[str] = ..., dbID: _Optional[int] = ..., collectionID: _Optional[int] = ..., partitionID: _Optional[int] = ..., schema: _Optional[bytes] = ..., virtualChannelNames: _Optional[_Iterable[str]] = ..., physicalChannelNames: _Optional[_Iterable[str]] = ..., partitionIDs: _Optional[_Iterable[int]] = ...) -> None: ...

class DropCollectionRequest(_message.Message):
    __slots__ = ("base", "db_name", "collectionName", "dbID", "collectionID")
    BASE_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONNAME_FIELD_NUMBER: _ClassVar[int]
    DBID_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    db_name: str
    collectionName: str
    dbID: int
    collectionID: int
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., db_name: _Optional[str] = ..., collectionName: _Optional[str] = ..., dbID: _Optional[int] = ..., collectionID: _Optional[int] = ...) -> None: ...

class CreatePartitionRequest(_message.Message):
    __slots__ = ("base", "db_name", "collection_name", "partition_name", "dbID", "collectionID", "partitionID")
    BASE_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_NAME_FIELD_NUMBER: _ClassVar[int]
    PARTITION_NAME_FIELD_NUMBER: _ClassVar[int]
    DBID_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    PARTITIONID_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    db_name: str
    collection_name: str
    partition_name: str
    dbID: int
    collectionID: int
    partitionID: int
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., db_name: _Optional[str] = ..., collection_name: _Optional[str] = ..., partition_name: _Optional[str] = ..., dbID: _Optional[int] = ..., collectionID: _Optional[int] = ..., partitionID: _Optional[int] = ...) -> None: ...

class DropPartitionRequest(_message.Message):
    __slots__ = ("base", "db_name", "collection_name", "partition_name", "dbID", "collectionID", "partitionID")
    BASE_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_NAME_FIELD_NUMBER: _ClassVar[int]
    PARTITION_NAME_FIELD_NUMBER: _ClassVar[int]
    DBID_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    PARTITIONID_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    db_name: str
    collection_name: str
    partition_name: str
    dbID: int
    collectionID: int
    partitionID: int
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., db_name: _Optional[str] = ..., collection_name: _Optional[str] = ..., partition_name: _Optional[str] = ..., dbID: _Optional[int] = ..., collectionID: _Optional[int] = ..., partitionID: _Optional[int] = ...) -> None: ...

class TimeTickMsg(_message.Message):
    __slots__ = ("base",)
    BASE_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ...) -> None: ...

class DataNodeTtMsg(_message.Message):
    __slots__ = ("base", "channel_name", "timestamp", "segments_stats")
    BASE_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    SEGMENTS_STATS_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    channel_name: str
    timestamp: int
    segments_stats: _containers.RepeatedCompositeFieldContainer[_common_pb2.SegmentStats]
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., channel_name: _Optional[str] = ..., timestamp: _Optional[int] = ..., segments_stats: _Optional[_Iterable[_Union[_common_pb2.SegmentStats, _Mapping]]] = ...) -> None: ...

class ReplicateMsg(_message.Message):
    __slots__ = ("base", "is_end", "is_cluster", "database", "collection")
    BASE_FIELD_NUMBER: _ClassVar[int]
    IS_END_FIELD_NUMBER: _ClassVar[int]
    IS_CLUSTER_FIELD_NUMBER: _ClassVar[int]
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    is_end: bool
    is_cluster: bool
    database: str
    collection: str
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., is_end: bool = ..., is_cluster: bool = ..., database: _Optional[str] = ..., collection: _Optional[str] = ...) -> None: ...

class ImportFile(_message.Message):
    __slots__ = ("id", "paths")
    ID_FIELD_NUMBER: _ClassVar[int]
    PATHS_FIELD_NUMBER: _ClassVar[int]
    id: int
    paths: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, id: _Optional[int] = ..., paths: _Optional[_Iterable[str]] = ...) -> None: ...

class ImportMsg(_message.Message):
    __slots__ = ("base", "db_name", "collection_name", "collectionID", "partitionIDs", "options", "files", "schema", "jobID")
    class OptionsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    BASE_FIELD_NUMBER: _ClassVar[int]
    DB_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONID_FIELD_NUMBER: _ClassVar[int]
    PARTITIONIDS_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    FILES_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    JOBID_FIELD_NUMBER: _ClassVar[int]
    base: _common_pb2.MsgBase
    db_name: str
    collection_name: str
    collectionID: int
    partitionIDs: _containers.RepeatedScalarFieldContainer[int]
    options: _containers.ScalarMap[str, str]
    files: _containers.RepeatedCompositeFieldContainer[ImportFile]
    schema: _schema_pb2.CollectionSchema
    jobID: int
    def __init__(self, base: _Optional[_Union[_common_pb2.MsgBase, _Mapping]] = ..., db_name: _Optional[str] = ..., collection_name: _Optional[str] = ..., collectionID: _Optional[int] = ..., partitionIDs: _Optional[_Iterable[int]] = ..., options: _Optional[_Mapping[str, str]] = ..., files: _Optional[_Iterable[_Union[ImportFile, _Mapping]]] = ..., schema: _Optional[_Union[_schema_pb2.CollectionSchema, _Mapping]] = ..., jobID: _Optional[int] = ...) -> None: ...



# hello_sprase.py demonstrates the basic operations of PyMilvus, a Python SDK of Milvus,
# while operating on sparse float vectors.
# 1. connect to Milvus
# 2. create collection
# 3. insert data
# 4. create index
# 5. search, query, and hybrid search on entities
# 6. delete entities by PK
# 7. drop collection
import time

import numpy as np
import random
from pymilvus import (
    connections,
    utility,
    FieldSchema, CollectionSchema, DataType,
    Collection,
)

fmt = "=== {:30} ==="
search_latency_fmt = "search latency = {:.4f}s"
num_entities, dim = 1000, 3000
# non zero count of randomly generated sparse vectors
nnz = 30

def log(msg):
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + " " + msg)

# -----------------------------------------------------------------------------
# connect to Milvus
log(fmt.format("start connecting to Milvus"))
connections.connect("default", host="localhost", port="19530")

has = utility.has_collection("hello_sparse")
log(f"Does collection hello_sparse exist in Milvus: {has}")

# -----------------------------------------------------------------------------
# create collection with a sparse float vector column
hello_sparse = None
if not has:
    fields = [
        FieldSchema(name="pk", dtype=DataType.VARCHAR, is_primary=True, auto_id=True, max_length=100),
        FieldSchema(name="random", dtype=DataType.DOUBLE),
        FieldSchema(name="embeddings", dtype=DataType.SPARSE_FLOAT_VECTOR),
    ]
    schema = CollectionSchema(fields, "hello_sparse is the simplest demo to introduce sparse float vector usage")
    log(fmt.format("Create collection `hello_sparse`"))
    hello_sparse = Collection("hello_sparse", schema, consistency_level="Strong")
else:
    hello_sparse = Collection("hello_sparse")

log(f"hello_sparse has {hello_sparse.num_entities} entities({hello_sparse.num_entities/1000000}M), indexed {hello_sparse.has_index()}")

# -----------------------------------------------------------------------------
# insert
log(fmt.format("Start creating entities to insert"))
rng = np.random.default_rng(seed=19530)

def generate_sparse_vector(dimension: int, non_zero_count: int) -> dict:
    indices = random.sample(range(dimension), non_zero_count)
    values = [random.random() for _ in range(non_zero_count)]
    sparse_vector = {index: value for index, value in zip(indices, values)}
    return sparse_vector

entities = [
    rng.random(num_entities).tolist(),
    [generate_sparse_vector(dim, nnz) for _ in range(num_entities)],
]

log(fmt.format("Start inserting entities"))
insert_result = hello_sparse.insert(entities)

# -----------------------------------------------------------------------------
# create index
if not hello_sparse.has_index():
    log(fmt.format("Start Creating index SPARSE_INVERTED_INDEX"))
    index = {
        "index_type": "SPARSE_INVERTED_INDEX",
        "metric_type": "IP",
        "params":{
            "drop_ratio_build": 0.2,
        }
    }
    hello_sparse.create_index("embeddings", index)

log(fmt.format("Start loading"))
hello_sparse.load()

# -----------------------------------------------------------------------------
# search based on vector similarity
log(fmt.format("Start searching based on vector similarity"))
vectors_to_search = entities[-1][-1:]
search_params = {
    "metric_type": "IP",
    "params": {
        "drop_ratio_search": "0.2",
    }
}

start_time = time.time()
result = hello_sparse.search(vectors_to_search, "embeddings", search_params, limit=3, output_fields=["pk", "random", "embeddings"])
end_time = time.time()

for hits in result:
    for hit in hits:
        print(f"hit: {hit}")
log(search_latency_fmt.format(end_time - start_time))
# -----------------------------------------------------------------------------
# query based on scalar filtering(boolean, int, etc.)
print(fmt.format("Start querying with `random > 0.5`"))

start_time = time.time()
result = hello_sparse.query(expr="random > 0.5", output_fields=["random", "embeddings"])
end_time = time.time()

print(f"query result:\n-{result[0]}")
print(search_latency_fmt.format(end_time - start_time))

# -----------------------------------------------------------------------------
# pagination
r1 = hello_sparse.query(expr="random > 0.5", limit=4, output_fields=["random"])
r2 = hello_sparse.query(expr="random > 0.5", offset=1, limit=3, output_fields=["random"])
print(f"query pagination(limit=4):\n\t{r1}")
print(f"query pagination(offset=1, limit=3):\n\t{r2}")


# -----------------------------------------------------------------------------
# hybrid search
print(fmt.format("Start hybrid searching with `random > 0.5`"))

start_time = time.time()
result = hello_sparse.search(vectors_to_search, "embeddings", search_params, limit=3, expr="random > 0.5", output_fields=["random"])
end_time = time.time()

for hits in result:
    for hit in hits:
        print(f"hit: {hit}, random field: {hit.entity.get('random')}")
print(search_latency_fmt.format(end_time - start_time))

# -----------------------------------------------------------------------------
# delete entities by PK
# You can delete entities by their PK values using boolean expressions.
ids = insert_result.primary_keys

expr = f'pk in ["{ids[0]}" , "{ids[1]}"]'
print(fmt.format(f"Start deleting with expr `{expr}`"))

result = hello_sparse.query(expr=expr, output_fields=["random", "embeddings"])
print(f"query before delete by expr=`{expr}` -> result: \n-{result[0]}\n-{result[1]}\n")

hello_sparse.delete(expr)

result = hello_sparse.query(expr=expr, output_fields=["random", "embeddings"])
print(f"query after delete by expr=`{expr}` -> result: {result}\n")


# -----------------------------------------------------------------------------
# drop collection
print(fmt.format("Drop collection `hello_sparse`"))
utility.drop_collection("hello_sparse")



import numpy as np
import time
import random
from pymilvus import (
    MilvusClient,
    DataType
)
from pymilvus.client.embedding_list import EmbeddingList

COLLECTION_NAME = "Documents"
EMBEDDING_DIM = 32
analyzer_params = {
    "type": "standard",
    "filter": ["lowercase"],
}
fmt = "\n=== {:30} ===\n"

milvus_client = MilvusClient("http://localhost:19530")
has_collection = milvus_client.has_collection(COLLECTION_NAME, timeout=5)
if has_collection:
    milvus_client.drop_collection(COLLECTION_NAME)

schema = milvus_client.create_schema(enable_dynamic_field=True, auto_id=True)
schema.add_field("id", DataType.INT64, is_primary=True, max_length=100)
schema.add_field("content", DataType.VARCHAR, max_length=10000, enable_analyzer=True,
                 analyzer_params=analyzer_params, enable_match=True)
schema.add_field("embedding", DataType.FLOAT_VECTOR, dim=16)


struct_schema = milvus_client.create_struct_field_schema()
struct_schema.add_field("struct_str", DataType.VARCHAR, max_length=65535)
struct_schema.add_field("struct_float_vec", DataType.FLOAT_VECTOR, dim=EMBEDDING_DIM)
# we can have multiple vector field in a struct
# struct_schema.add_field("struct_float16_vec", DataType.FLOAT16_VECTOR, dim=EMBEDDING_DIM)
schema.add_field("struct_field",datatype=DataType.ARRAY, element_type=DataType.STRUCT, struct_schema=struct_schema, max_capacity=1000)

struct_schema = milvus_client.create_struct_field_schema()
struct_schema.add_field("struct_str", DataType.VARCHAR, max_length=65535)
struct_schema.add_field("struct_float_vec", DataType.FLOAT_VECTOR, dim=EMBEDDING_DIM)
schema.add_field("struct_field2",datatype=DataType.ARRAY, element_type=DataType.STRUCT, struct_schema=struct_schema, max_capacity=1000)


milvus_client.create_collection(COLLECTION_NAME, schema=schema)

index_params = milvus_client.prepare_index_params()
dense_index_params = {
    "index_type": "IVF_FLAT",  # Choose an appropriate index type (e.g., IVF_FLAT)
    "metric_type": "COSINE",   # Metric type (e.g., COSINE for cosine similarity)
    "params": {"nlist": 128}   # Index parameters
}

index_params.add_index(field_name="embedding", index_type="IVF_FLAT", metric_type="COSINE", index_params={"nlist": 128})
index_params.add_index(field_name="struct_field[struct_float_vec]", index_type="HNSW", index_name="struct_float_vec_index1", metric_type="MAX_SIM_COSINE", index_params={"M": 16, "efConstruction": 200})
# index_params.add_index(field_name="struct_field[struct_float16_vec]", index_type="HNSW", index_name="struct_float_vec_index2",metric_type="MAX_SIM_COSINE", index_params={"M": 16, "efConstruction": 200})
index_params.add_index(field_name="struct_field2[struct_float_vec]", index_type="HNSW", index_name="struct_float_vec_index3", metric_type="MAX_SIM_IP", index_params={"M": 16, "efConstruction": 200})
milvus_client.create_index(COLLECTION_NAME, schema=schema, index_params=index_params)

coll = milvus_client.describe_collection(COLLECTION_NAME)
print("Describe collection", coll)

rng = np.random.default_rng(seed=19530)

N = 100
content = ["aaa",
           "jjj",
           "sss",
           "asdasd",
           "dvzicxv",
           "dcmxvxv",
           "dvxcnvmlzc",
           "aczxcc",
           "qiqixcc"]
iterations = 10
for _ in range(iterations):
    arr_len = random.randint(2, 5)
    rows = [{"embedding": rng.random((1, 16))[0],
             "content": content[random.randint(0, len(content) - 1)],
             "struct_field": [
                 {
                  "struct_str": content[random.randint(0, len(content) - 1)],
                  "struct_float_vec": rng.random((1, EMBEDDING_DIM))[0],
                  # "struct_float16_vec": rng.random((1, EMBEDDING_DIM))[0].astype(np.float16),
                  }
                  for _ in range(arr_len)],
             "struct_field2": [
                 {
                     "struct_str": content[random.randint(0, len(content) - 1)],
                     "struct_float_vec": rng.random((1, EMBEDDING_DIM))[0],
                 }
                 for _ in range(arr_len)],
             }
            for _ in range(N)]
    result = milvus_client.insert(COLLECTION_NAME, rows)

milvus_client.flush(COLLECTION_NAME)

milvus_client.load_collection(COLLECTION_NAME)

result = milvus_client.query(
    collection_name=COLLECTION_NAME,
    filter="",
    output_fields=["struct_field"],
    limit=2,
)

print("Query", result)

rng = np.random.default_rng(seed=19530)



# Create search queries using EmbeddingList
# For testing purposes, using random test data
queries = [
    EmbeddingList._from_random_test(7, EMBEDDING_DIM, seed=19530),  # Query with 7 vectors
    EmbeddingList._from_random_test(4, EMBEDDING_DIM, seed=19531),  # Query with 4 vectors
]

embeddingList = EmbeddingList()
# Query with 2 vectors
embeddingList.add(np.random.randn(EMBEDDING_DIM))
embeddingList.add(np.random.randn(EMBEDDING_DIM))

queries.append(embeddingList)

field = "struct_field[struct_float_vec]"
res = milvus_client.search(COLLECTION_NAME, data=queries, limit=10, anns_field=field,
                     output_fields=["struct_field[struct_str]"])

for i, (hits, query) in enumerate(zip(res, queries)):
    print(f"============== Query {i+1} ({query}) - Total hits: {len(hits)}")
    for hit in hits:
        print(f"    Hit id: {hit.id}, distance: {hit.distance}")
        print(f"        entity: {hit}")


import asyncio
import time
import numpy as np
from pymilvus import AsyncMilvusClient, DataType

fmt = "\n=== {:30} ===\n"
dim = 128
collection_name = "test_collection"
test_user = "test_user"
test_role = "test_role"
test_alias = "test_alias"
test_db = "test_database"
partition_name = "test_partition"
group_name = "test_privilege_group"
rg1_name = "test_resource_group"
rg2_name = "test_resource_group"

async def create_resources(client):
    print(fmt.format("Creating Resources"))

    print("Creating database...")
    await client.create_database(test_db)
    print(f"Database {test_db} created")

    print("Creating user...")
    await client.create_user(test_user, "password123")
    print(f"User {test_user} created")

    print("Creating privilege group...")
    await client.create_privilege_group(group_name)
    print(f"Privilege group {group_name} created")

    print("Creating collection...")
    schema = client.create_schema(
        auto_id=True, description="Test collection", enable_dynamic_field=True
    )
    schema.add_field("id", DataType.INT64, is_primary=True, auto_id=True)
    schema.add_field("vector", DataType.FLOAT_VECTOR, dim=dim)
    schema.add_field("text", DataType.VARCHAR, max_length=512)

    index_params = client.prepare_index_params()
    index_params.add_index("vector", index_type="HNSW", metric_type="L2")

    await client.create_collection(collection_name, schema=schema)
    print(f"Collection {collection_name} created")
    
    print("Creating partition...")
    await client.create_partition(collection_name, partition_name)
    print(f"Partition {partition_name} created")
    
    await client.create_index(collection_name, index_params=index_params)
    print("Index created")
    
    print("Creating alias...")
    await client.create_alias(collection_name, test_alias)
    print(f"Alias {test_alias} created")

    print("Creating resource groups...")
    await client.create_resource_group(rg1_name)
    print(f"Resource group {rg1_name} created")
    await client.create_resource_group(rg2_name)
    print(f"Resource group {rg2_name} created")
    
    print("Loading collection...")
    await client.load_collection(collection_name)
    print("Collection loaded")
    
    return True

async def test_functionality(client):
    print(fmt.format("Testing Functionality"))

    print("Testing server info...")
    server_version = await client.get_server_version()
    print(f"Server version: {server_version}")

    print("Testing collection operations...")
    print(f"has_collection: {await client.has_collection(collection_name)}")
    print(f"list_collections: {await client.list_collections()}")
    print(f"describe_collection: {await client.describe_collection(collection_name)}")

    await client.rename_collection(collection_name, f"{collection_name}_renamed")
    await client.rename_collection(f"{collection_name}_renamed", collection_name)
    print("Collection rename test completed")

    await client.alter_collection_properties(collection_name, {"collection.ttl.seconds": 3600})
    await client.drop_collection_properties(collection_name, ["collection.ttl.seconds"])
    print("Collection properties test completed")

    collection_info = await client.describe_collection(collection_name)
    enable_dynamic = collection_info.get('enable_dynamic_field', False)
    print(f"Dynamic fields enabled: {enable_dynamic}")

    await client.release_collection(collection_name)
    await client.alter_collection_field(collection_name, "text", {"max_length": 256})
    print("Collection field altered")
    await client.load_collection(collection_name)
    print("Collection reloaded")

    print("Testing index operations...")
    print(f"list_indexes: {await client.list_indexes(collection_name)}")
    index_name = (await client.list_indexes(collection_name))[0]
    print(f"describe_index: {await client.describe_index(collection_name, index_name)}")

    await client.release_collection(collection_name)
    await client.alter_index_properties(collection_name, index_name, {"mmap.enabled": True})
    await client.drop_index_properties(collection_name, index_name, ["mmap.enabled"])
    print("Index properties test completed")
    await client.load_collection(collection_name)
    print("Collection reloaded after index operations")

    print("Testing partition operations...")
    print(f"has_partition: {await client.has_partition(collection_name, partition_name)}")
    print(f"list_partitions: {await client.list_partitions(collection_name)}")

    print("Testing user management...")
    print(f"list_users: {await client.list_users()}")
    print(f"describe_user: {await client.describe_user(test_user)}")

    await client.update_password(test_user, "password123", "newpassword123")
    print("Password updated")

    print("Testing alias operations...")
    print(f"describe_alias: {await client.describe_alias(test_alias)}")
    print(f"list_aliases: {await client.list_aliases(collection_name)}")

    await client.alter_alias(collection_name, test_alias)
    print("Alias altered")

    print("Testing data operations...")
    data = [{
        "vector": np.random.random(dim).tolist(),
        "text": f"test text {i}",
        "dynamic_field_int": i * 10,
        "dynamic_field_str": f"dynamic_{i}",
        "dynamic_field_float": i * 0.5
    } for i in range(10)]

    result = await client.insert(collection_name, data)
    print(f"insert result: {result}")

    await client.flush(collection_name)
    print("flush completed")

    compaction_id = await client.compact(collection_name)
    print(f"compact job id: {compaction_id}")
    print(f"compaction state: {await client.get_compaction_state(compaction_id)}")

    print("Testing stats and load state...")
    print(f"get_collection_stats: {await client.get_collection_stats(collection_name)}")
    partition_stats = await client.get_partition_stats(collection_name, partition_name)
    print(f"get_partition_stats: {partition_stats}")
    print(f"get_load_state: {await client.get_load_state(collection_name)}")

    await client.refresh_load(collection_name)
    print(f"describe_replica: {await client.describe_replica(collection_name)}")

    print("Testing text analysis...")
    analyzer_result = await client.run_analyzer(
        texts=["hello world", "test text"],
        analyzer_params={"tokenizer": "standard"}
    )
    print(f"analyzer result: {analyzer_result}")

    print("Testing database operations...")
    print(f"list_databases: {await client.list_databases()}")
    print(f"describe_database: {await client.describe_database(test_db)}")

    client.use_database(test_db)
    client.using_database("default")
    print("Database switch test completed")

    await client.alter_database_properties(test_db, {"database.replica.num": 1})
    await client.drop_database_properties(test_db, ["database.replica.num"])
    print("Database properties test completed")

    print("Testing privilege group operations...")
    print(f"list_privilege_groups: {await client.list_privilege_groups()}")

    await client.add_privileges_to_group(group_name, ["Insert", "Delete"])
    await client.remove_privileges_from_group(group_name, ["Delete"])
    print("Privilege group operations completed")

    print(fmt.format("Testing create_field_schema"))
    field1 = client.create_field_schema("test_id", DataType.INT64, desc="test int64 field", is_primary=True, auto_id=True)
    print(f"Field1: {field1}")
    field2 = client.create_field_schema("test_vector", DataType.FLOAT_VECTOR, desc="test vector field", dim=dim)
    print(f"Field2: {field2}")
    field3 = client.create_field_schema("test_text", DataType.VARCHAR, desc="test varchar field", max_length=256)
    print(f"Field3: {field3}")
    del field1
    del field2
    del field3
    print("create_field_schema test completed")

    print(fmt.format("Testing Resource Group Operations"))

    # List resource groups
    print("\nListing resource groups...")
    rgs = await client.list_resource_groups()
    print(f"Resource groups: {rgs}")
    assert rg1_name in rgs
    assert rg2_name in rgs

    # Describe resource groups
    print(f"\nDescribing resource group {rg1_name}...")
    rg1_info = await client.describe_resource_group(rg1_name)
    print(f"Resource group {rg1_name} info: {rg1_info}")
    
    print(f"\nDescribing resource group {rg2_name}...")
    rg2_info = await client.describe_resource_group(rg2_name)
    print(f"Resource group {rg2_name} info: {rg2_info}")

async def cleanup_resources(client):
    print(fmt.format("Cleaning Up Resources"))

    print("Dropping alias...")
    await client.drop_alias(test_alias)
    print(f"Alias {test_alias} dropped")

    print("Dropping collection...")
    await client.drop_collection(collection_name)
    print(f"Collection {collection_name} dropped")

    print("Dropping user...")
    await client.drop_user(test_user)
    print(f"User {test_user} dropped")

    print("Dropping privilege group...")
    await client.drop_privilege_group(group_name)
    print(f"Privilege group {group_name} dropped")

    print("Dropping database...")
    await client.drop_database(test_db)
    print(f"Database {test_db} dropped")

    print("\nDropping resource groups...")
    if rg1_name in await client.list_resource_groups():
        await client.drop_resource_group(rg1_name)
        print(f"Resource group {rg1_name} dropped")
    if rg2_name in await client.list_resource_groups():
        await client.drop_resource_group(rg2_name)
        print(f"Resource group {rg2_name} dropped")

async def main():
    client = AsyncMilvusClient("http://localhost:19530")
    await create_resources(client)
    await test_functionality(client)
    time.sleep(10)
    await cleanup_resources(client)
    time.sleep(10)
    await client.close()
    print(fmt.format("Test Completed"))

if __name__ == "__main__":
    asyncio.run(main())


import time
import numpy as np
from pymilvus import (
    MilvusClient,
    DataType,
    Function,
    FunctionType,
    AnnSearchRequest,
)

fmt = "\n=== {:30} ===\n"
dim = 8
collection_name = "hello_milvus"
milvus_client = MilvusClient("http://localhost:19530")

has_collection = milvus_client.has_collection(collection_name, timeout=5)
if has_collection:
    milvus_client.drop_collection(collection_name)

schema = milvus_client.create_schema(enable_dynamic_field=False, auto_id=True)
schema.add_field("id", DataType.INT64, is_primary=True)
schema.add_field("embeddings", DataType.FLOAT_VECTOR, dim=dim)
schema.add_field("ts", DataType.INT64)


index_params = milvus_client.prepare_index_params()
index_params.add_index(field_name = "embeddings", index_type="FLAT", metric_type="L2")
milvus_client.create_collection(collection_name, schema=schema, index_params=index_params, consistency_level="Strong")

print(fmt.format("    all collections    "))
print(milvus_client.list_collections())

print(fmt.format(f"schema of collection {collection_name}"))
print(milvus_client.describe_collection(collection_name))

rng = np.random.default_rng(seed=19530)
rows = [
        {"embeddings": rng.random((1, dim))[0], "ts": 100},
        {"embeddings": rng.random((1, dim))[0], "ts": 200},
        {"embeddings": rng.random((1, dim))[0], "ts": 300},
        {"embeddings": rng.random((1, dim))[0], "ts": 400},
        {"embeddings": rng.random((1, dim))[0], "ts": 500},
        {"embeddings": rng.random((1, dim))[0], "ts": 600},
]

print(fmt.format("Start inserting entities"))
insert_result = milvus_client.insert(collection_name, rows)
print(fmt.format("Inserting entities done"))
print(insert_result)


print(fmt.format("Start load collection "))
milvus_client.load_collection(collection_name)

rng = np.random.default_rng(seed=19530)
vectors_to_search = rng.random((1, dim))

ranker = Function(
    name="rerank_fn",
    input_field_names=["ts"],
    function_type=FunctionType.RERANK,
    params={
        "reranker": "decay",
        "function": "exp",
        "origin": 0,
        "offset": 200,
        "decay": 0.9,
        "scale": 100
    }
)

print(fmt.format(f"Start search with retrieve serveral fields."))
result = milvus_client.search(collection_name, vectors_to_search, limit=3, output_fields=["*"], ranker=ranker)
for hits in result:
    for hit in hits:
        print(f"hit: {hit}")

vectors_to_search = rng.random((1, dim))
search_param = {
    "data": vectors_to_search,
    "anns_field": "embeddings",
    "param": {"metric_type": "L2"},
    "limit": 3,
}
req = AnnSearchRequest(**search_param)

hybrid_res = milvus_client.hybrid_search(collection_name, [req, req], ranker=ranker, limit=3, output_fields=["ts"])
for hits in hybrid_res:
    for hit in hits:
        print(f" hybrid search hit: {hit}")

milvus_client.drop_collection(collection_name)



import logging

from pymilvus.bulk_writer.stage_restful import create_stage, delete_stage, list_stages

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StageManager:
    def __init__(self, cloud_endpoint: str, api_key: str):
        """
        private preview feature. Please submit a request and contact us if you need it.

        Args:
            cloud_endpoint (str): The fixed cloud endpoint URL.
                - For international regions: https://api.cloud.zilliz.com
                - For regions in China: https://api.cloud.zilliz.com.cn
            api_key (str): The API key associated with your organization or cluster.
        """
        self.cloud_endpoint = cloud_endpoint
        self.api_key = api_key

    def create_stage(self, project_id: str, region_id: str, stage_name: str):
        """
        Create a stage under the specified project and regionId.
        """
        create_stage(self.cloud_endpoint, self.api_key, project_id, region_id, stage_name)

    def delete_stage(self, stage_name: str):
        """
        Delete a stage.
        """
        delete_stage(self.cloud_endpoint, self.api_key, stage_name)

    def list_stages(self, project_id: str, current_page: int = 1, page_size: int = 10):
        """
        Paginated query of the stage list under a specified projectId.
        """
        return list_stages(self.cloud_endpoint, self.api_key, project_id, current_page, page_size)



"""EmbeddingList: A container for multiple embeddings for array-of-vector searches."""

from typing import Any, List, Optional, Union

import numpy as np

from pymilvus.client.types import DataType
from pymilvus.exceptions import ParamError


class EmbeddingList:
    """
    A container for multiple embeddings that can be used directly in Milvus searches.
    Represents a single query containing multiple vectors for array-of-vector fields.

    This is particularly useful for searching struct array fields that contain vectors,
    enabling array-of-vector to array-of-vector searches.

    Examples:
        >>> # Create empty and add vectors
        >>> query1 = EmbeddingList()
        >>> query1.add(embedding1)
        >>> query1.add(embedding2)
        >>>
        >>> # Create from list of vectors
        >>> vectors = [vec1, vec2, vec3]
        >>> query2 = EmbeddingList(vectors)
        >>>
        >>> # Use in search
        >>> results = client.search(
        >>>     collection_name="my_collection",
        >>>     data=[query1, query2],  # List of EmbeddingList
        >>>     ...
        >>> )
    """

    def __init__(
        self,
        embeddings: Optional[Union[List[np.ndarray], np.ndarray]] = None,
        dim: Optional[int] = None,
        dtype: Optional[Union[np.dtype, str, DataType]] = None,
    ):
        """
        Initialize an EmbeddingList.

        Args:
            embeddings: Initial embeddings. Can be:
                - None: Creates an empty list
                - np.ndarray with shape (n, dim): Batch of n embeddings
                - np.ndarray with shape (dim,): Single embedding
                - List[np.ndarray]: List of embedding arrays
            dim: Expected dimension for validation (optional).
                 If provided, all added embeddings will be validated against this dimension.
            dtype: Data type of the embeddings (optional). Can be:
                - numpy dtype (e.g., np.float32, np.float16, np.uint8)
                - string (e.g., 'float32', 'float16', 'uint8')
                - DataType enum (e.g., DataType.FLOAT_VECTOR, DataType.BINARY_VECTOR)
                If not specified, will infer from the first embedding added.
        """
        self._embeddings: List[np.ndarray] = []
        self._dim = dim
        self._dtype = self._parse_dtype(dtype) if dtype is not None else None

        if embeddings is not None:
            if isinstance(embeddings, np.ndarray):
                if embeddings.ndim == 1:
                    # Single vector
                    self.add(embeddings)
                elif embeddings.ndim == 2:
                    # Multiple vectors
                    for i in range(len(embeddings)):
                        self.add(embeddings[i])
                else:
                    msg = "Embeddings array must be 1D or 2D"
                    raise ValueError(msg)
            elif isinstance(embeddings, list):
                for emb in embeddings:
                    self.add(emb)
            else:
                msg = "Embeddings must be numpy array or list"
                raise TypeError(msg)

    def _parse_dtype(self, dtype: Union[np.dtype, str, DataType]) -> np.dtype:
        """Parse and validate data type."""
        if isinstance(dtype, np.dtype):
            return dtype
        if isinstance(dtype, str):
            return np.dtype(dtype)
        if isinstance(dtype, DataType):
            # Map DataType enum to numpy dtype
            dtype_map = {
                DataType.FLOAT_VECTOR: np.float32,
                DataType.FLOAT16_VECTOR: np.float16,
                DataType.BFLOAT16_VECTOR: np.float16,  # Use float16 as approximation
                DataType.BINARY_VECTOR: np.uint8,
                DataType.INT8_VECTOR: np.int8,
            }
            if dtype in dtype_map:
                return np.dtype(dtype_map[dtype])
            msg = f"Unsupported DataType: {dtype}"
            raise ParamError(message=msg)
        msg = f"dtype must be numpy dtype, string, or DataType, got {type(dtype)}"
        raise TypeError(msg)

    def _infer_dtype(self, array: np.ndarray) -> np.dtype:
        """Infer dtype from array, with smart defaults."""
        if array.dtype == np.float64:
            # Default double precision to single precision for efficiency
            return np.dtype(np.float32)
        return array.dtype

    def add(self, embedding: Union[np.ndarray, List[Any]]) -> "EmbeddingList":
        """
        Add a single embedding vector to the list.

        Args:
            embedding: A single embedding vector (1D array or list)

        Returns:
            Self for method chaining

        Raises:
            ValueError: If embedding dimension doesn't match existing embeddings
        """
        embedding = np.asarray(embedding)

        if embedding.ndim != 1:
            msg = f"Embedding must be 1D, got shape {embedding.shape}"
            raise ValueError(msg)

        # Validate dimension
        if self._embeddings:
            if len(embedding) != self.dim:
                msg = f"Embedding dimension {len(embedding)} doesn't match existing {self.dim}"
                raise ValueError(msg)
        elif self._dim is not None and len(embedding) != self._dim:
            msg = f"Embedding dimension {len(embedding)} doesn't match expected {self._dim}"
            raise ValueError(msg)

        # Handle dtype
        if self._dtype is None:
            # Infer dtype from first embedding
            self._dtype = self._infer_dtype(embedding)

        # Convert to the established dtype
        if embedding.dtype != self._dtype:
            embedding = embedding.astype(self._dtype)

        self._embeddings.append(embedding)
        return self

    def add_batch(self, embeddings: Union[List[np.ndarray], np.ndarray]) -> "EmbeddingList":
        """
        Add multiple embeddings at once.

        Args:
            embeddings: Batch of embeddings (2D array or list of 1D arrays)

        Returns:
            Self for method chaining

        Raises:
            ValueError: If embeddings have inconsistent dimensions
        """
        if isinstance(embeddings, np.ndarray):
            if embeddings.ndim != 2:
                msg = f"Batch embeddings must be 2D, got {embeddings.ndim}D"
                raise ValueError(msg)
            for i in range(len(embeddings)):
                self.add(embeddings[i])
        else:
            for emb in embeddings:
                self.add(emb)
        return self

    @classmethod
    def _from_random_test(
        cls,
        num_vectors: int,
        dim: int,
        dtype: Optional[Union[np.dtype, str, DataType]] = None,
        seed: Optional[int] = None,
    ) -> "EmbeddingList":
        """
        Create an EmbeddingList with random vectors for testing purposes.

        WARNING: This method is intended for testing and demonstration only.
        Do not use in production code.

        Args:
            num_vectors: Number of random vectors to generate
            dim: Dimension of each vector
            dtype: Data type of the vectors (default: np.float32)
            seed: Random seed for reproducibility

        Returns:
            New EmbeddingList with random test vectors
        """
        rng = np.random.default_rng(seed)

        # Default dtype to float32 if None
        if dtype is None:
            dtype = np.dtype(np.float32)

        # Parse dtype if needed
        if not isinstance(dtype, np.dtype):
            dtype = cls(None, dim=dim, dtype=dtype)._dtype

        # Generate random data based on dtype
        if dtype == np.uint8:
            # For binary vectors, generate random bits
            embeddings = rng.integers(0, 256, size=(num_vectors, dim), dtype=np.uint8)
        elif dtype == np.int8:
            # For int8 vectors
            embeddings = rng.integers(-128, 128, size=(num_vectors, dim), dtype=np.int8)
        elif dtype in [np.float16, np.float32, np.float64]:
            # For float vectors
            embeddings = rng.random((num_vectors, dim)).astype(dtype)
        else:
            msg = f"Unsupported dtype for random generation: {dtype}"
            raise ValueError(msg)

        return cls(embeddings, dim=dim, dtype=dtype)

    def to_flat_array(self) -> np.ndarray:
        """
        Convert to flat array format required by Milvus for array-of-vector fields.

        Returns:
            Flattened numpy array containing all embeddings concatenated

        Raises:
            ValueError: If the list is empty
        """
        if not self._embeddings:
            msg = "EmbeddingList is empty"
            raise ValueError(msg)
        # Preserve the dtype of the embeddings
        return np.concatenate(self._embeddings)

    def to_numpy(self) -> np.ndarray:
        """
        Convert to 2D numpy array.

        Returns:
            2D numpy array with shape (num_embeddings, dim)

        Raises:
            ValueError: If the list is empty
        """
        if not self._embeddings:
            msg = "EmbeddingList is empty"
            raise ValueError(msg)
        return np.stack(self._embeddings)

    def clear(self) -> "EmbeddingList":
        """Clear all embeddings from the list."""
        self._embeddings.clear()
        return self

    def __len__(self) -> int:
        """Return the number of embeddings in the list."""
        return len(self._embeddings)

    def __getitem__(self, index: int) -> np.ndarray:
        """Get embedding at specific index."""
        return self._embeddings[index]

    def __iter__(self):
        """Iterate over embeddings."""
        return iter(self._embeddings)

    @property
    def dim(self) -> int:
        """Dimension of each embedding."""
        if self._embeddings:
            return len(self._embeddings[0])
        return self._dim or 0

    @property
    def dtype(self) -> Optional[np.dtype]:
        """Data type of the embeddings."""
        return self._dtype

    @property
    def shape(self) -> tuple:
        """Shape as (num_embeddings, dim)."""
        return (len(self), self.dim)

    @property
    def total_dim(self) -> int:
        """Total dimension of all embeddings combined."""
        return len(self) * self.dim

    @property
    def is_empty(self) -> bool:
        """Check if the list is empty."""
        return len(self._embeddings) == 0

    @property
    def nbytes(self) -> int:
        """Total number of bytes used by all embeddings."""
        if self.is_empty:
            return 0
        return sum(emb.nbytes for emb in self._embeddings)

    def __repr__(self) -> str:
        dtype_str = f", dtype={self.dtype}" if self.dtype else ""
        return f"EmbeddingList(count={len(self)}, dim={self.dim}{dtype_str})"

    def __str__(self) -> str:
        if self.is_empty:
            return "EmbeddingList(empty)"
        dtype_str = f" ({self.dtype})" if self.dtype else ""
        return f"EmbeddingList with {len(self)} embeddings of dimension {self.dim}{dtype_str}"



from pymilvus import (
    MilvusClient,
    DataType,
)
from pymilvus.client.constants import DEFAULT_RESOURCE_GROUP

from pymilvus.client.types import (
    ResourceGroupConfig,
)

fmt = "\n=== {:30} ===\n"
dim = 8
collection_name = "hello_milvus"
milvus_client = MilvusClient("http://localhost:19530")


## create collection and load collection
print("create collection and load collection")
collection_name = "hello_milvus"
has_collection = milvus_client.has_collection(collection_name, timeout=5)
if has_collection:
    milvus_client.drop_collection(collection_name)

schema = milvus_client.create_schema(enable_dynamic_field=True)
schema.add_field("id", DataType.INT64, is_primary=True)
schema.add_field("embeddings", DataType.FLOAT_VECTOR, dim=dim)
schema.add_field("title", DataType.VARCHAR, max_length=64)
milvus_client.create_collection(collection_name, schema=schema, consistency_level="Strong")
index_params = milvus_client.prepare_index_params()
index_params.add_index(field_name = "embeddings", metric_type="L2")
index_params.add_index(field_name = "title", index_type = "Trie", index_name="my_trie")
milvus_client.create_index(collection_name, index_params)
milvus_client.load_collection(collection_name)


## create resource group
print("create resource group")
milvus_client.create_resource_group("rg1")
milvus_client.create_resource_group("rg2")

## update resource group
configs = {
            "rg1": ResourceGroupConfig(
                requests={"node_num": 1},
                limits={"node_num": 5},
                transfer_from=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
                transfer_to=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
            ),
            "rg2": ResourceGroupConfig(
                requests={"node_num": 4},
                limits={"node_num": 4},
                transfer_from=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
                transfer_to=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
            ),
        }
milvus_client.update_resource_groups(configs)

## describe resource group
print("describe rg1")
result = milvus_client.describe_resource_group("rg1")
print(result)

print("describe rg2")
result = milvus_client.describe_resource_group("rg2")
print(result)

## list resource group
print("list resource group")
result = milvus_client.list_resource_groups()
print(result)

## transfer replica
print("transfer replica to rg1")
milvus_client.transfer_replica(DEFAULT_RESOURCE_GROUP, "rg1", collection_name, 1)
print("describe rg1 after transfer replica in")
result = milvus_client.describe_resource_group("rg1")
print(result)

milvus_client.transfer_replica("rg1", DEFAULT_RESOURCE_GROUP, collection_name, 1)
print("describe rg1 after transfer replica out")
result = milvus_client.describe_resource_group("rg1")
print(result)

## drop resource group
print("drop resource group")
# create resource group
configs = {
            "rg1": ResourceGroupConfig(
                requests={"node_num": 0},
                limits={"node_num": 0},
                transfer_from=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
                transfer_to=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
            ),
            "rg2": ResourceGroupConfig(
                requests={"node_num": 0},
                limits={"node_num": 0},
                transfer_from=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
                transfer_to=[{"resource_group": DEFAULT_RESOURCE_GROUP}],
            ),
        }
milvus_client.update_resource_groups(configs)
milvus_client.drop_resource_group("rg1")
milvus_client.drop_resource_group("rg2")





from . import common_pb2 as _common_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ResourceGroupLimit(_message.Message):
    __slots__ = ("node_num",)
    NODE_NUM_FIELD_NUMBER: _ClassVar[int]
    node_num: int
    def __init__(self, node_num: _Optional[int] = ...) -> None: ...

class ResourceGroupTransfer(_message.Message):
    __slots__ = ("resource_group",)
    RESOURCE_GROUP_FIELD_NUMBER: _ClassVar[int]
    resource_group: str
    def __init__(self, resource_group: _Optional[str] = ...) -> None: ...

class ResourceGroupNodeFilter(_message.Message):
    __slots__ = ("node_labels",)
    NODE_LABELS_FIELD_NUMBER: _ClassVar[int]
    node_labels: _containers.RepeatedCompositeFieldContainer[_common_pb2.KeyValuePair]
    def __init__(self, node_labels: _Optional[_Iterable[_Union[_common_pb2.KeyValuePair, _Mapping]]] = ...) -> None: ...

class ResourceGroupConfig(_message.Message):
    __slots__ = ("requests", "limits", "transfer_from", "transfer_to", "node_filter")
    REQUESTS_FIELD_NUMBER: _ClassVar[int]
    LIMITS_FIELD_NUMBER: _ClassVar[int]
    TRANSFER_FROM_FIELD_NUMBER: _ClassVar[int]
    TRANSFER_TO_FIELD_NUMBER: _ClassVar[int]
    NODE_FILTER_FIELD_NUMBER: _ClassVar[int]
    requests: ResourceGroupLimit
    limits: ResourceGroupLimit
    transfer_from: _containers.RepeatedCompositeFieldContainer[ResourceGroupTransfer]
    transfer_to: _containers.RepeatedCompositeFieldContainer[ResourceGroupTransfer]
    node_filter: ResourceGroupNodeFilter
    def __init__(self, requests: _Optional[_Union[ResourceGroupLimit, _Mapping]] = ..., limits: _Optional[_Union[ResourceGroupLimit, _Mapping]] = ..., transfer_from: _Optional[_Iterable[_Union[ResourceGroupTransfer, _Mapping]]] = ..., transfer_to: _Optional[_Iterable[_Union[ResourceGroupTransfer, _Mapping]]] = ..., node_filter: _Optional[_Union[ResourceGroupNodeFilter, _Mapping]] = ...) -> None: ...



import copy

import numpy as np
import pandas as pd
import pytest
from pymilvus import CollectionSchema, DataType, FieldSchema
from pymilvus.orm.schema import Function, FunctionType


class TestCollectionSchema:
    @pytest.fixture(scope="function")
    def raw_dict(self):
        return {
            "description": "TestCollectionSchema_description",
            "enable_dynamic_field": True,
            "fields": [
                {
                    "name": "vec1",
                    "description": "desc1",
                    "type": DataType.FLOAT_VECTOR,
                    "params": {"dim": 128},
                },
                {
                    "name": "vec2",
                    "description": "desc2",
                    "type": DataType.BINARY_VECTOR,
                    "params": {"dim": 128},
                },
                {
                    "name": "ID",
                    "description": "ID",
                    "type": DataType.INT64,
                    "is_primary": True,
                    "auto_id": False
                },
            ]
        }

    def test_constructor_from_dict(self, raw_dict):
        schema = CollectionSchema.construct_from_dict(raw_dict)
        assert schema.enable_dynamic_field == raw_dict.get("enable_dynamic_field", False)
        assert schema.description, raw_dict['description']
        assert len(schema.fields) == len(raw_dict['fields'])
        f = schema.primary_field
        assert isinstance(f, FieldSchema)
        assert f.name == raw_dict['fields'][2]['name']

    def test_to_dict(self, raw_dict):
        schema = CollectionSchema.construct_from_dict(raw_dict)
        target = schema.to_dict()
        target.pop("auto_id", None)
        assert target == raw_dict
        assert target is not raw_dict

    def test_init_with_functions(self, raw_dict):
        functions = [
            Function("func1", FunctionType.BM25, ["field1"], ["field2"])
        ]
        schema = CollectionSchema.construct_from_dict(raw_dict)
        schema_with_func = CollectionSchema(schema.fields, schema.description, functions=functions)
        assert schema_with_func.functions == functions


class TestFieldSchema:
    @pytest.fixture(scope="function")
    def raw_dict_float_vector(self):
        return  {
            "name": "TestFieldSchema_name_floatvector",
            "description": "TestFieldSchema_description_floatvector",
            "type": DataType.FLOAT_VECTOR,
            "params": {"dim": 128},
        }

    @pytest.fixture(scope="function")
    def raw_dict_binary_vector(self):
        return {
            "name": "TestFieldSchema_name_binary_vector",
            "description": "TestFieldSchema_description_binary_vector",
            "type": DataType.BINARY_VECTOR,
            "params": {"dim": 128},
        }

    @pytest.fixture(scope="function")
    def raw_dict_float16_vector(self):
        return {
            "name": "TestFieldSchema_name_float16_vector",
            "description": "TestFieldSchema_description_float16_vector",
            "type": DataType.FLOAT16_VECTOR,
            "params": {"dim": 128},
        }

    @pytest.fixture(scope="function")
    def raw_dict_bfloat16_vector(self):
        return {
            "name": "TestFieldSchema_name_bfloat16_vector",
            "description": "TestFieldSchema_description_bfloat16_vector",
            "type": DataType.BFLOAT16_VECTOR,
            "params": {"dim": 128},
        }

    @pytest.fixture(scope="function")
    def raw_dict_int8_vector(self):
        return {
            "name": "TestFieldSchema_name_int8_vector",
            "description": "TestFieldSchema_description_int8_vector",
            "type": DataType.INT8_VECTOR,
            "params": {"dim": 128},
        }

    @pytest.fixture(scope="function")
    def raw_dict_norm(self):
        return {
            "name": "TestFieldSchema_name_norm",
            "description": "TestFieldSchema_description_norm",
            "type": DataType.INT64,
        }

    @pytest.fixture(scope="function")
    def dataframe1(self):
        data = {
            'float': [1.0],
            'int32': [2],
            'float_vec': [np.array([3, 4.0], np.float32)]
        }
        return pd.DataFrame(data)

    def test_constructor_from_float_dict(self, raw_dict_float_vector):
        field = FieldSchema.construct_from_dict(raw_dict_float_vector)
        assert field.dtype == DataType.FLOAT_VECTOR
        assert field.description == raw_dict_float_vector['description']
        assert field.is_primary is False
        assert field.name == raw_dict_float_vector['name']
        assert field.dim == raw_dict_float_vector['params']['dim']

    def test_constructor_from_binary_dict(self, raw_dict_binary_vector):
        field = FieldSchema.construct_from_dict(raw_dict_binary_vector)
        assert field.dtype == DataType.BINARY_VECTOR
        assert field.description == raw_dict_binary_vector['description']
        assert field.is_primary is False
        assert field.name == raw_dict_binary_vector['name']
        assert field.dim == raw_dict_binary_vector['params']['dim']

    def test_constructor_from_float16_dict(self, raw_dict_float16_vector):
        field = FieldSchema.construct_from_dict(raw_dict_float16_vector)
        assert field.dtype == DataType.FLOAT16_VECTOR
        assert field.description == raw_dict_float16_vector['description']
        assert field.is_primary is False
        assert field.name == raw_dict_float16_vector['name']
        assert field.dim == raw_dict_float16_vector['params']['dim']

    def test_constructor_from_bfloat16_dict(self, raw_dict_bfloat16_vector):
        field = FieldSchema.construct_from_dict(raw_dict_bfloat16_vector)
        assert field.dtype == DataType.BFLOAT16_VECTOR
        assert field.description == raw_dict_bfloat16_vector['description']
        assert field.is_primary is False
        assert field.name == raw_dict_bfloat16_vector['name']
        assert field.dim == raw_dict_bfloat16_vector['params']['dim']

    def test_constructor_from_int8_dict(self, raw_dict_int8_vector):
        field = FieldSchema.construct_from_dict(raw_dict_int8_vector)
        assert field.dtype == DataType.INT8_VECTOR
        assert field.description == raw_dict_int8_vector['description']
        assert field.is_primary is False
        assert field.name == raw_dict_int8_vector['name']
        assert field.dim == raw_dict_int8_vector['params']['dim']

    def test_constructor_from_norm_dict(self, raw_dict_norm):
        field = FieldSchema.construct_from_dict(raw_dict_norm)
        assert field.dtype == DataType.INT64
        assert field.description == raw_dict_norm['description']
        assert field.is_primary is False
        assert field.name == raw_dict_norm['name']
        assert field.dim is None
        assert field.dummy is None

    def test_cmp(self, raw_dict_binary_vector):
        field1 = FieldSchema.construct_from_dict(raw_dict_binary_vector)
        field2 = FieldSchema.construct_from_dict(raw_dict_binary_vector)
        assert field1 == field2
        dict1 = copy.deepcopy(raw_dict_binary_vector)
        dict1["name"] = dict1["name"] + "_"
        field3 = FieldSchema.construct_from_dict(dict1)
        assert field1 != field3

    def test_to_dict(self, raw_dict_norm, raw_dict_float_vector, raw_dict_binary_vector):
        fields = []
        dicts = [raw_dict_norm, raw_dict_float_vector, raw_dict_binary_vector]
        fields.append(FieldSchema.construct_from_dict(raw_dict_norm))
        fields.append(FieldSchema.construct_from_dict(raw_dict_float_vector))
        fields.append(FieldSchema.construct_from_dict(raw_dict_binary_vector))

        for i, f in enumerate(fields):
            target = f.to_dict()
            assert target == dicts[i]
            assert target is not dicts[i]



import numpy as np
from pymilvus import (
    MilvusClient,
    DataType,
)

fmt = "\n=== {:30} ===\n"
search_latency_fmt = "search latency = {:.4f}s"
num_entities, dim = 3000, 8

collection_name = "hello_milvus"
milvus_client = MilvusClient("http://localhost:19530")

has_collection = milvus_client.has_collection(collection_name, timeout=5)
if has_collection:
    milvus_client.drop_collection(collection_name)

schema = milvus_client.create_schema(auto_id=False, description="hello_milvus is the simplest demo to introduce the APIs")
schema.add_field("pk", DataType.VARCHAR, is_primary=True, max_length=100)
schema.add_field("random", DataType.DOUBLE)
schema.add_field("embeddings", DataType.FLOAT_VECTOR, dim=dim)

index_params = milvus_client.prepare_index_params()
index_params.add_index(field_name = "embeddings", index_type = "IVF_FLAT", metric_type="L2", nlist=128)

print(fmt.format("Create collection `hello_milvus`"))

milvus_client.create_collection(collection_name, schema=schema, index_params=index_params, consistency_level="Strong")


print(fmt.format("Start inserting entities"))
rng = np.random.default_rng(seed=19530)
entities = [
    # provide the pk field because `auto_id` is set to False
    [str(i) for i in range(num_entities)],
    rng.random(num_entities).tolist(),  # field random, only supports list
    rng.random((num_entities, dim)),    # field embeddings, supports numpy.ndarray and list
]

rows = [ {"pk": entities[0][i], "random": entities[1][i], "embeddings": entities[2][i]} for i in range (num_entities)]

insert_result = milvus_client.insert(collection_name, rows)


print(fmt.format("Start loading"))
milvus_client.load_collection(collection_name)

field_name = "embeddings"

req_list = []
nq = 1
default_limit = 5
vectors_to_search = []

filters = {
    "pk == {str}": {"str": "10"},
    "pk in {list}": {"list": ["1", "10", "100"]},
    "random > {target}": {"target": 5},
    "random <= {target}": {"target": 111.5},
    "{min} <= random < {max}": {"min": 0, "max": 9999},
}

search_param = {
    "data": vectors_to_search,
    "anns_field": field_name,
    "param": {"metric_type": "L2"},
    "limit": default_limit}
vectors_to_search = rng.random((nq, dim))

for filter, filter_params in filters.items():
    print(f"search with filter: {filter}")
    result = milvus_client.search(collection_name=collection_name, data=vectors_to_search, filter=filter, limit=3,
                                  output_fields=["random"], search_params=search_param, filter_params=filter_params)

    for hits in result:
        for hit in hits:
            print(f"hit: {hit}")

    query_results = milvus_client.query(collection_name, filter=filter, output_fields=["random"],
                                        filter_params=filter_params, limit=3)
    for ret in query_results:
        print("query result: ", ret)

print(fmt.format("Search after delete"))
ids = insert_result["ids"]
filter = "pk in {list}"
filter_params = {"list": [ids[0], ids[1]]}
print(f"Start deleting with filter `{filter}`")

result = milvus_client.query(collection_name, filter=filter, output_fields=["random"],
                             filter_params=filter_params, limit=3)
print(f"query before delete by filter=`{filter}` -> result: \n-{result[0]}\n-{result[1]}\n")

milvus_client.delete(collection_name=collection_name, filter=filter, filter_params=filter_params)

result = milvus_client.query(collection_name, filter=filter, output_fields=["random"],
                             filter_params=filter_params, limit=3)
print(f"query after delete by filter=`{filter}` -> result: {result}\n")

print(fmt.format("Release collection"))
milvus_client.release_collection(collection_name)

print(fmt.format("Drop collection"))
milvus_client.drop_collection(collection_name)



from pymilvus import (
    connections,
    FieldSchema, CollectionSchema, DataType,
    Collection,
    utility,
    Role
)

# This example shows how to:
#   1.  connect to Milvus server
#   2.  create a role
#   3.  add user to the role
#   4.  assign the role to the user
#   5.  create privilege group
#   6.  add privileges to the privilege group
#   7.  remove privileges from the privilege group
#   8.  list privilege groups of the role
#   9.  grant the privilege group to the role
#   10. grant a built-in privilege group to the role
#   11. revoke the privilege group from the role
#   12. revoke a built-in privilege group from the role
#   13. drop the role

_HOST = 'localhost'
_USER = 'root'
_PASSWORD = 'Milvus'

# Create a Milvus connection
def create_connection():
    print(f"\nCreate connection...")
    connections.connect(host=_HOST, user=_USER, password=_PASSWORD)
    print(f"\nList connections:")
    print(connections.list_connections())

def create_privilege_group(role, group_name):
    role.create_privilege_group(group_name)

def add_privileges_to_group(role, group_name, privileges):
    role.add_privileges_to_group(group_name, privileges)

def list_privilege_groups(role):
    return role.list_privilege_groups()

def remove_privileges_from_group(role, group_name, privileges):
    role.remove_privileges_from_group(group_name, privileges)

def drop_privilege_group(role, group_name):
    role.drop_privilege_group(group_name)

def grant_v2(role, privilege, collection_name, db_name):
    role.grant_v2(privilege, collection_name, db_name=db_name)

def revoke_v2(role, privilege, collection_name, db_name):
    role.revoke_v2(privilege, collection_name, db_name=db_name)

def list_grants(role):
    return role.list_grants()

def main():
    # Connect to Milvus server
    create_connection()

    # create role
    role = Role("privilege_group_role")
    role.create()

    # list roles
    utility.list_roles(True)

    # create user
    utility.create_user(user="user1", password="Milvus")

    # add user
    role.add_user("user1")

    # create privilege group
    privilege_group = "search_query"
    create_privilege_group(role, privilege_group)

    # add privileges to group
    add_privileges_to_group(role, privilege_group, ["Search", "Query"])

    # list privilege groups
    groups = list_privilege_groups(role)
    print("List privilege groups: ", groups)

    # remove privileges from group
    remove_privileges_from_group(role, privilege_group, ["Query"])

    # grant custom privielge group to role
    grant_v2(role, privilege_group, "*", db_name="*")

    # grant built-in privielge group to role
    grant_v2(role, "ClusterReadOnly", "*", db_name="*")

    # list grants
    grants =list_grants(role)
    print("List grants: ", grants)

    # revoke custom privielge group from role
    revoke_v2(role, privilege_group, "*", db_name="*")

    # revoke built-in privielge group from role
    revoke_v2(role, "ClusterReadOnly", "*", db_name="*")

    # drop privilege group
    drop_privilege_group(role, privilege_group)

    # drop role
    role.drop()


if __name__ == '__main__':
    main()



import logging.config
import os

from dotenv import load_dotenv

load_dotenv()


class Config:
    # legacy env MILVUS_DEFAULT_CONNECTION, not recommended
    LEGACY_URI = str(os.getenv("MILVUS_DEFAULT_CONNECTION", ""))
    MILVUS_URI = str(os.getenv("MILVUS_URI", LEGACY_URI))

    MILVUS_CONN_ALIAS = str(os.getenv("MILVUS_CONN_ALIAS", "default"))
    MILVUS_CONN_TIMEOUT = float(os.getenv("MILVUS_CONN_TIMEOUT", "10.0"))

    # legacy configs:
    DEFAULT_USING = MILVUS_CONN_ALIAS
    DEFAULT_CONNECT_TIMEOUT = MILVUS_CONN_TIMEOUT

    # TODO tidy the following configs
    GRPC_PORT = "19530"
    GRPC_ADDRESS = "127.0.0.1:19530"
    GRPC_URI = f"tcp://{GRPC_ADDRESS}"

    DEFAULT_HOST = "localhost"
    DEFAULT_PORT = "19530"

    WaitTimeDurationWhenLoad = 0.2  # in seconds
    MaxVarCharLengthKey = "max_length"
    MaxVarCharLength = 65535
    EncodeProtocol = "utf-8"
    IndexName = ""


# logging
COLORS = {
    "HEADER": "\033[95m",
    "INFO": "\033[92m",
    "DEBUG": "\033[94m",
    "WARNING": "\033[93m",
    "ERROR": "\033[95m",
    "CRITICAL": "\033[91m",
    "ENDC": "\033[0m",
}


class ColorFulFormatColMixin:
    def format_col(self, message_str: str, level_name: str):
        if level_name in COLORS:
            message_str = COLORS.get(level_name) + message_str + COLORS.get("ENDC")
        return message_str


class ColorfulFormatter(logging.Formatter, ColorFulFormatColMixin):
    def format(self, record: str):
        message_str = super().format(record)

        return self.format_col(message_str, level_name=record.levelname)


def init_log(log_level: str):
    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "format": "%(asctime)s [%(levelname)s][%(funcName)s]: %(message)s (%(filename)s:%(lineno)s)",
            },
            "colorful_console": {
                "format": "%(asctime)s | %(levelname)s: %(message)s (%(filename)s:%(lineno)s) (%(process)s)",
                "()": ColorfulFormatter,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "colorful_console",
            },
            "no_color_console": {
                "class": "logging.StreamHandler",
                "formatter": "default",
            },
        },
        "loggers": {
            "pymilvus": {"handlers": ["no_color_console"], "level": log_level, "propagate": False},
            "pymilvus.milvus_client": {
                "handlers": ["no_color_console"],
                "level": "INFO",
                "propagate": False,
            },
            "pymilvus.bulk_writer": {
                "handlers": ["no_color_console"],
                "level": "INFO",
                "propagate": False,
            },
        },
    }
    logging.config.dictConfig(logging_config)


init_log("WARNING")



import logging
import math
import struct
from typing import Any, Dict, Iterable, List, Optional

import numpy as np
import orjson

from pymilvus.exceptions import (
    DataNotMatchException,
    ExceptionsMessage,
    MilvusException,
    ParamError,
)
from pymilvus.grpc_gen import schema_pb2 as schema_types
from pymilvus.settings import Config

from .types import DataType
from .utils import (
    SciPyHelper,
    SparseMatrixInputType,
    SparseRowOutputType,
    sparse_parse_single_row,
)

logger = logging.getLogger(__name__)

CHECK_STR_ARRAY = True


def entity_is_sparse_matrix(entity: Any):
    if SciPyHelper.is_scipy_sparse(entity):
        return True
    try:

        def is_type_in_str(v: Any, t: Any):
            if not isinstance(v, str):
                return False
            try:
                t(v)
            except ValueError:
                return False
            return True

        def is_int_type(v: Any):
            return isinstance(v, (int, np.integer)) or is_type_in_str(v, int)

        def is_float_type(v: Any):
            return isinstance(v, (float, np.floating)) or is_type_in_str(v, float)

        # must be of multiple rows
        if len(entity) == 0:
            return False
        for item in entity:
            if SciPyHelper.is_scipy_sparse(item):
                return item.shape[0] == 1
            if not isinstance(item, dict) and not isinstance(item, list):
                return False
            pairs = item.items() if isinstance(item, dict) else item
            # each row must be a list of Tuple[int, float]. we allow empty sparse row
            for pair in pairs:
                if len(pair) != 2 or not is_int_type(pair[0]) or not is_float_type(pair[1]):
                    return False
    except Exception:
        return False
    return True


# converts supported sparse matrix to schemapb.SparseFloatArray proto
def sparse_rows_to_proto(data: SparseMatrixInputType) -> schema_types.SparseFloatArray:
    # converts a sparse float vector to plain bytes. the format is the same as how
    # milvus interprets/persists the data.
    def sparse_float_row_to_bytes(indices: Iterable[int], values: Iterable[float]):
        if len(indices) != len(values):
            raise ParamError(
                message=f"length of indices and values must be the same, got {len(indices)} and {len(values)}"
            )
        data = b""
        for i, v in sorted(zip(indices, values), key=lambda x: x[0]):
            if not (0 <= i < 2**32 - 1):
                raise ParamError(
                    message=f"sparse vector index must be positive and less than 2^32-1: {i}"
                )
            if math.isnan(v):
                raise ParamError(message="sparse vector value must not be NaN")
            data += struct.pack("I", i)
            data += struct.pack("f", v)
        return data

    if not entity_is_sparse_matrix(data):
        raise ParamError(message="input must be a sparse matrix in supported format")

    result = schema_types.SparseFloatArray()

    if SciPyHelper.is_scipy_sparse(data):
        csr = data.tocsr()
        result.dim = csr.shape[1]
        for start, end in zip(csr.indptr[:-1], csr.indptr[1:]):
            result.contents.append(
                sparse_float_row_to_bytes(csr.indices[start:end], csr.data[start:end])
            )
    else:
        dim = 0
        for _, row_data in enumerate(data):
            if SciPyHelper.is_scipy_sparse(row_data):
                if row_data.shape[0] != 1:
                    raise ParamError(message="invalid input for sparse float vector: expect 1 row")
                dim = max(dim, row_data.shape[1])
                result.contents.append(sparse_float_row_to_bytes(row_data.indices, row_data.data))
            else:
                indices = []
                values = []
                row = row_data.items() if isinstance(row_data, dict) else row_data
                for index, value in row:
                    indices.append(int(index))
                    values.append(float(value))
                result.contents.append(sparse_float_row_to_bytes(indices, values))
                row_dim = 0
                if len(indices) > 0:
                    row_dim = indices[-1] + 1
                dim = max(dim, row_dim)
        result.dim = dim
    return result


# converts schema_types.SparseFloatArray proto to Iterable[SparseRowOutputType]
def sparse_proto_to_rows(
    sfv: schema_types.SparseFloatArray, start: Optional[int] = None, end: Optional[int] = None
) -> Iterable[SparseRowOutputType]:
    if not isinstance(sfv, schema_types.SparseFloatArray):
        raise ParamError(message="Vector must be a sparse float vector")
    start = start or 0
    end = end or len(sfv.contents)
    return [sparse_parse_single_row(row_bytes) for row_bytes in sfv.contents[start:end]]


def get_input_num_rows(entity: Any) -> int:
    if SciPyHelper.is_scipy_sparse(entity):
        return entity.shape[0]
    return len(entity)


def entity_type_to_dtype(entity_type: Any):
    if isinstance(entity_type, int):
        return entity_type
    if isinstance(entity_type, str):
        # case sensitive
        return schema_types.DataType.Value(entity_type)
    raise ParamError(message=f"invalid entity type: {entity_type}")


def get_max_len_of_var_char(field_info: Dict) -> int:
    k = Config.MaxVarCharLengthKey
    v = Config.MaxVarCharLength
    return field_info.get("params", {}).get(k, v)


def convert_to_str_array(orig_str_arr: Any, field_info: Dict, check: bool = True):
    arr = []
    if Config.EncodeProtocol.lower() != "utf-8".lower():
        for s in orig_str_arr:
            arr.append(s.encode(Config.EncodeProtocol))
    else:
        arr = orig_str_arr
    max_len = int(get_max_len_of_var_char(field_info))
    if check:
        for s in arr:
            if not isinstance(s, str):
                raise ParamError(
                    message=f"field ({field_info['name']}) expects string input, got: {type(s)}"
                )
            if len(s) > max_len:
                raise ParamError(
                    message=f"invalid input of field ({field_info['name']}), "
                    f"length of string exceeds max length. length: {len(s)}, max length: {max_len}"
                )
    return arr


def entity_to_str_arr(entity_values: Any, field_info: Any, check: bool = True):
    return convert_to_str_array(entity_values, field_info, check=check)


def convert_to_json(obj: object):
    def preprocess_numpy_types(obj: Any):
        if isinstance(obj, dict):
            return {k: preprocess_numpy_types(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [preprocess_numpy_types(item) for item in obj]
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.bool_):
            return bool(obj)
        return obj

    # Handle JSON string input
    if isinstance(obj, str):
        try:
            # Validate JSON string by parsing it
            parsed_obj = orjson.loads(obj)
            # If it's a valid JSON string, validate dict keys if it's a dict
            if isinstance(parsed_obj, dict):
                for k in parsed_obj:
                    if not isinstance(k, str):
                        raise DataNotMatchException(message=ExceptionsMessage.JSONKeyMustBeStr)
            # Return the original string encoded as bytes (since it's already valid JSON)
            return obj.encode(Config.EncodeProtocol)
        except Exception as e:
            # Truncate the string if it's too long for better readability
            max_len = 200
            json_str_display = obj if len(obj) <= max_len else obj[:max_len] + "..."
            raise DataNotMatchException(
                message=f"Invalid JSON string: {e!s}. Input string: {json_str_display!r}"
            ) from e

    # Handle dict input
    if isinstance(obj, dict):
        for k in obj:
            if not isinstance(k, str):
                raise DataNotMatchException(message=ExceptionsMessage.JSONKeyMustBeStr)

    processed_obj = preprocess_numpy_types(obj)

    return orjson.dumps(processed_obj)


def convert_to_json_arr(objs: List[object], field_info: Any):
    arr = []
    for obj in objs:
        if obj is None:
            raise ParamError(message=f"field ({field_info['name']}) expects a non-None input")
        arr.append(convert_to_json(obj))
    return arr


def entity_to_json_arr(entity_values: Dict, field_info: Any):
    return convert_to_json_arr(entity_values, field_info)


def convert_to_array_arr(objs: List[Any], field_info: Any):
    return [convert_to_array(obj, field_info) for obj in objs]


def convert_to_array(obj: List[Any], field_info: Any):
    field_data = schema_types.ScalarField()
    element_type = field_info.get("element_type", None)
    if element_type == DataType.BOOL:
        field_data.bool_data.data.extend(obj)
        return field_data
    if element_type in (DataType.INT8, DataType.INT16, DataType.INT32):
        field_data.int_data.data.extend(obj)
        return field_data
    if element_type == DataType.INT64:
        field_data.long_data.data.extend(obj)
        return field_data
    if element_type == DataType.FLOAT:
        field_data.float_data.data.extend(obj)
        return field_data
    if element_type == DataType.DOUBLE:
        field_data.double_data.data.extend(obj)
        return field_data
    if element_type in (DataType.VARCHAR, DataType.STRING):
        field_data.string_data.data.extend(obj)
        return field_data
    raise ParamError(
        message=f"Unsupported element type: {element_type} for Array field: {field_info.get('name')}"
    )


def convert_to_array_of_vector(obj: List[Any], field_info: Any):
    # Create a single VectorField that contains all vectors flattened
    field_data = schema_types.VectorField()
    element_type = field_info.get("element_type", None)

    dim_value = field_info.get("params", {}).get("dim", 0)
    if isinstance(dim_value, str):
        dim_value = int(dim_value)
    field_data.dim = dim_value

    if element_type == DataType.FLOAT_VECTOR:
        if not obj:
            field_data.float_vector.data.extend([])
        for field_value in obj:
            f_value = field_value
            if isinstance(field_value, np.ndarray):
                if field_value.dtype not in ("float32", "float64"):
                    raise ParamError(
                        message="invalid input for float32 vector. Expected an np.ndarray with dtype=float32"
                    )
                f_value = field_value.tolist()
            field_data.float_vector.data.extend(f_value)

    else:
        # todo(SpadeA): other types are now not supported. When it's supported, make sure empty
        # array is handled correctly.
        raise ParamError(
            message=f"Unsupported element type: {element_type} for Array of Vector field: {field_info.get('name')}"
        )
    return field_data


def entity_to_array_arr(entity_values: List[Any], field_info: Any):
    return convert_to_array_arr(entity_values, field_info)


def pack_field_value_to_field_data(
    field_value: Any, field_data: schema_types.FieldData, field_info: Any
):
    field_type = field_data.type
    field_name = field_info["name"]
    if field_type == DataType.BOOL:
        try:
            if field_value is None:
                field_data.scalars.bool_data.data.extend([])
            else:
                field_data.scalars.bool_data.data.append(field_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "bool", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type in (DataType.INT8, DataType.INT16, DataType.INT32):
        try:
            # need to extend it, or cannot correctly identify field_data.scalars.int_data.data
            if field_value is None:
                field_data.scalars.int_data.data.extend([])
            else:
                field_data.scalars.int_data.data.append(field_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "int", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.INT64:
        try:
            if field_value is None:
                field_data.scalars.long_data.data.extend([])
            else:
                field_data.scalars.long_data.data.append(field_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "int64", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.FLOAT:
        try:
            if field_value is None:
                field_data.scalars.float_data.data.extend([])
            else:
                field_data.scalars.float_data.data.append(field_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "float", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.DOUBLE:
        try:
            if field_value is None:
                field_data.scalars.double_data.data.extend([])
            else:
                field_data.scalars.double_data.data.append(field_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "double", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.FLOAT_VECTOR:
        try:
            f_value = field_value
            if isinstance(field_value, np.ndarray):
                if field_value.dtype not in ("float32", "float64"):
                    raise ParamError(
                        message="invalid input for float32 vector. Expected an np.ndarray with dtype=float32"
                    )
                f_value = field_value.tolist()

            field_data.vectors.dim = len(f_value)
            field_data.vectors.float_vector.data.extend(f_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "float_vector", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.BINARY_VECTOR:
        try:
            field_data.vectors.dim = len(field_value) * 8
            field_data.vectors.binary_vector += bytes(field_value)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "binary_vector", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.FLOAT16_VECTOR:
        try:
            if isinstance(field_value, bytes):
                v_bytes = field_value
            elif isinstance(field_value, np.ndarray):
                if field_value.dtype != "float16":
                    raise ParamError(
                        message="invalid input for float16 vector. Expected an np.ndarray with dtype=float16"
                    )
                v_bytes = field_value.view(np.uint8).tobytes()
            else:
                raise ParamError(
                    message="invalid input type for float16 vector. Expected an np.ndarray with dtype=float16"
                )

            field_data.vectors.dim = len(v_bytes) // 2
            field_data.vectors.float16_vector += v_bytes
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "float16_vector", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.BFLOAT16_VECTOR:
        try:
            if isinstance(field_value, bytes):
                v_bytes = field_value
            elif isinstance(field_value, np.ndarray):
                if field_value.dtype != "bfloat16":
                    raise ParamError(
                        message="invalid input for bfloat16 vector. Expected an np.ndarray with dtype=bfloat16"
                    )
                v_bytes = field_value.view(np.uint8).tobytes()
            else:
                raise ParamError(
                    message="invalid input type for bfloat16 vector. Expected an np.ndarray with dtype=bfloat16"
                )

            field_data.vectors.dim = len(v_bytes) // 2
            field_data.vectors.bfloat16_vector += v_bytes
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "bfloat16_vector", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.SPARSE_FLOAT_VECTOR:
        try:
            if not SciPyHelper.is_scipy_sparse(field_value):
                field_value = [field_value]
            elif field_value.shape[0] != 1:
                raise ParamError(message="invalid input for sparse float vector: expect 1 row")
            if not entity_is_sparse_matrix(field_value):
                raise ParamError(message="invalid input for sparse float vector")
            field_data.vectors.sparse_float_vector.contents.append(
                sparse_rows_to_proto(field_value).contents[0]
            )
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "sparse_float_vector", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.INT8_VECTOR:
        try:
            if isinstance(field_value, np.ndarray):
                if field_value.dtype != "int8":
                    raise ParamError(
                        message="invalid input for int8 vector. Expected an np.ndarray with dtype=int8"
                    )
                i_bytes = field_value.view(np.int8).tobytes()
            else:
                raise ParamError(
                    message="invalid input for int8 vector. Expected an np.ndarray with dtype=int8"
                )

            field_data.vectors.dim = len(i_bytes)
            field_data.vectors.int8_vector += i_bytes
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "int8_vector", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.VARCHAR:
        try:
            if field_value is None:
                field_data.scalars.string_data.data.extend([])
            else:
                field_data.scalars.string_data.data.append(
                    convert_to_str_array(field_value, field_info, CHECK_STR_ARRAY)
                )
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "varchar", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.GEOMETRY:
        try:
            if field_value is None:
                field_data.scalars.geometry_wkt_data.data.extend([])
            else:
                field_data.scalars.geometry_wkt_data.data.append(
                    convert_to_str_array(field_value, field_info, CHECK_STR_ARRAY)
                )
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "geometry", type(field_value))
            ) from e
    elif field_type == DataType.JSON:
        try:
            if field_value is None:
                field_data.scalars.json_data.data.extend([])
            else:
                field_data.scalars.json_data.data.append(convert_to_json(field_value))
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "json", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    elif field_type == DataType.ARRAY:
        try:
            if field_value is None:
                field_data.scalars.array_data.data.extend([])
            else:
                field_data.scalars.array_data.data.append(convert_to_array(field_value, field_info))
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(
                message=ExceptionsMessage.FieldDataInconsistent
                % (field_name, "array", type(field_value))
                + f" Detail: {e!s}"
            ) from e
    else:
        raise ParamError(message=f"Unsupported data type: {field_type}")


# Don't change entity inside.
def entity_to_field_data(entity: Dict, field_info: Any, num_rows: int) -> schema_types.FieldData:
    entity_type = entity.get("type")
    field_name = entity.get("name")
    entity_values = entity.get("values")
    field_data = schema_types.FieldData(
        type=entity_type_to_dtype(entity_type),
        field_name=field_name,
    )

    valid_data = []
    if field_info.get("nullable", False) or field_info.get("default_value", None):
        if len(entity_values) == 0:
            valid_data = [False] * num_rows
        else:
            valid_data = [value is not None for value in entity_values]
            entity_values = [value for value in entity_values if value is not None]

    field_data.valid_data.extend(valid_data)

    try:
        if entity_type == DataType.BOOL:
            field_data.scalars.bool_data.data.extend(entity_values)
        elif entity_type in (DataType.INT8, DataType.INT16, DataType.INT32):
            field_data.scalars.int_data.data.extend(entity_values)
        elif entity_type == DataType.INT64:
            field_data.scalars.long_data.data.extend(entity_values)
        elif entity_type == DataType.FLOAT:
            field_data.scalars.float_data.data.extend(entity_values)
        elif entity_type == DataType.DOUBLE:
            field_data.scalars.double_data.data.extend(entity_values)

        elif entity_type == DataType.FLOAT_VECTOR:
            # TODO: get dimension from field_info
            field_data.vectors.dim = len(entity_values[0])
            all_floats = [f for vector in entity_values for f in vector]
            field_data.vectors.float_vector.data.extend(all_floats)
        elif entity_type == DataType.BINARY_VECTOR:
            field_data.vectors.dim = len(entity_values[0]) * 8
            field_data.vectors.binary_vector = b"".join(entity_values)
        elif entity_type == DataType.FLOAT16_VECTOR:
            field_data.vectors.dim = len(entity_values[0]) // 2
            field_data.vectors.float16_vector = b"".join(entity_values)
        elif entity_type == DataType.BFLOAT16_VECTOR:
            field_data.vectors.dim = len(entity_values[0]) // 2
            field_data.vectors.bfloat16_vector = b"".join(entity_values)
        elif entity_type == DataType.SPARSE_FLOAT_VECTOR:
            field_data.vectors.sparse_float_vector.CopyFrom(sparse_rows_to_proto(entity_values))
        elif entity_type == DataType.INT8_VECTOR:
            field_data.vectors.dim = len(entity_values[0])
            field_data.vectors.int8_vector = b"".join(entity_values)

        elif entity_type == DataType.VARCHAR:
            field_data.scalars.string_data.data.extend(
                entity_to_str_arr(entity_values, field_info, CHECK_STR_ARRAY)
            )
        elif entity_type == DataType.JSON:
            field_data.scalars.json_data.data.extend(entity_to_json_arr(entity_values, field_info))
        elif entity_type == DataType.ARRAY:
            field_data.scalars.array_data.data.extend(
                entity_to_array_arr(entity_values, field_info)
            )
        elif entity_type == DataType.GEOMETRY:
            field_data.scalars.geometry_wkt_data.data.extend(
                entity_to_str_arr(entity_values, field_info, CHECK_STR_ARRAY)
            )
        else:
            raise ParamError(message=f"Unsupported data type: {entity_type}")
    except (TypeError, ValueError) as e:
        raise DataNotMatchException(
            message=ExceptionsMessage.FieldDataInconsistent
            % (field_name, entity_type.name, type(entity_values[0]))
            + f" Detail: {e!s}"
        ) from e
    return field_data


def extract_dynamic_field_from_result(raw: Any):
    dynamic_field_name = None
    field_names = set()
    if raw.fields_data:
        for field_data in raw.fields_data:
            field_names.add(field_data.field_name)
            if field_data.is_dynamic:
                dynamic_field_name = field_data.field_name

    dynamic_fields = set()
    for name in raw.output_fields:
        if name == dynamic_field_name:
            dynamic_fields.clear()
            break
        if name not in field_names:
            dynamic_fields.add(name)
    return dynamic_field_name, dynamic_fields


def extract_array_row_data_with_validity(field_data: Any, entity_rows: List[Dict], row_count: int):
    field_name = field_data.field_name
    data = field_data.scalars.array_data.data
    element_type = field_data.scalars.array_data.element_type
    if element_type == DataType.INT64:
        [
            entity_rows[i].__setitem__(
                field_name, data[i].long_data.data if field_data.valid_data[i] else None
            )
            for i in range(row_count)
        ]
    elif element_type == DataType.BOOL:
        [
            entity_rows[i].__setitem__(
                field_name, data[i].bool_data.data if field_data.valid_data[i] else None
            )
            for i in range(row_count)
        ]
    elif element_type in (
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
    ):
        [
            entity_rows[i].__setitem__(
                field_name, data[i].int_data.data if field_data.valid_data[i] else None
            )
            for i in range(row_count)
        ]
    elif element_type == DataType.FLOAT:
        [
            entity_rows[i].__setitem__(
                field_name, data[i].float_data.data if field_data.valid_data[i] else None
            )
            for i in range(row_count)
        ]
    elif element_type == DataType.DOUBLE:
        [
            entity_rows[i].__setitem__(
                field_name, data[i].double_data.data if field_data.valid_data[i] else None
            )
            for i in range(row_count)
        ]
    elif element_type in (
        DataType.STRING,
        DataType.VARCHAR,
    ):
        [
            entity_rows[i].__setitem__(
                field_name, data[i].string_data.data if field_data.valid_data[i] else None
            )
            for i in range(row_count)
        ]
    else:
        raise MilvusException(message=f"Unsupported data type: {element_type}")


def extract_array_row_data_no_validity(field_data: Any, entity_rows: List[Dict], row_count: int):
    field_name = field_data.field_name
    data = field_data.scalars.array_data.data
    element_type = field_data.scalars.array_data.element_type
    if element_type == DataType.INT64:
        [entity_rows[i].__setitem__(field_name, data[i].long_data.data) for i in range(row_count)]
    elif element_type == DataType.BOOL:
        [entity_rows[i].__setitem__(field_name, data[i].bool_data.data) for i in range(row_count)]
    elif element_type in (
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
    ):
        [entity_rows[i].__setitem__(field_name, data[i].int_data.data) for i in range(row_count)]
    elif element_type == DataType.FLOAT:
        [entity_rows[i].__setitem__(field_name, data[i].float_data.data) for i in range(row_count)]
    elif element_type == DataType.DOUBLE:
        [entity_rows[i].__setitem__(field_name, data[i].double_data.data) for i in range(row_count)]
    elif element_type in (
        DataType.STRING,
        DataType.VARCHAR,
    ):
        [entity_rows[i].__setitem__(field_name, data[i].string_data.data) for i in range(row_count)]
    else:
        raise MilvusException(message=f"Unsupported data type: {element_type}")


def extract_array_row_data(field_data: Any, index: int):
    array = field_data.scalars.array_data.data[index]
    if field_data.scalars.array_data.element_type == DataType.INT64:
        return array.long_data.data
    if field_data.scalars.array_data.element_type == DataType.BOOL:
        return array.bool_data.data
    if field_data.scalars.array_data.element_type in (
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
    ):
        return array.int_data.data
    if field_data.scalars.array_data.element_type == DataType.FLOAT:
        return array.float_data.data
    if field_data.scalars.array_data.element_type == DataType.DOUBLE:
        return array.double_data.data
    if field_data.scalars.array_data.element_type in (
        DataType.STRING,
        DataType.VARCHAR,
    ):
        return array.string_data.data
    return None


def extract_row_data_from_fields_data_v2(
    field_data: Any,
    entity_rows: List[Dict],
) -> bool:
    row_count = len(entity_rows)
    has_valid = len(field_data.valid_data) > 0
    field_name = field_data.field_name
    valid_data = field_data.valid_data

    def assign_scalar(data: List[Any]) -> None:
        if has_valid:
            [
                entity_rows[i].__setitem__(field_name, None if not valid_data[i] else data[i])
                for i in range(row_count)
            ]
        else:
            [entity_rows[i].__setitem__(field_name, data[i]) for i in range(row_count)]

    if field_data.type == DataType.BOOL:
        data = field_data.scalars.bool_data.data
        assign_scalar(data)
        return False

    if field_data.type in (DataType.INT8, DataType.INT16, DataType.INT32):
        data = field_data.scalars.int_data.data
        assign_scalar(data)
        return False

    if field_data.type == DataType.INT64:
        data = field_data.scalars.long_data.data
        assign_scalar(data)
        return False

    if field_data.type == DataType.FLOAT:
        data = field_data.scalars.float_data.data
        assign_scalar(data)
        return False

    if field_data.type == DataType.DOUBLE:
        data = field_data.scalars.double_data.data
        assign_scalar(data)
        return False

    if field_data.type == DataType.VARCHAR:
        data = field_data.scalars.string_data.data
        assign_scalar(data)
        return False

    if field_data.type == DataType.GEOMETRY:
        data = field_data.scalars.geometry_wkt_data.data
        assign_scalar(data)
        return False

    if field_data.type == DataType.JSON:
        return True

    if field_data.type == DataType.ARRAY:
        if has_valid:
            extract_array_row_data_with_validity(field_data, entity_rows, row_count)
        else:
            extract_array_row_data_no_validity(field_data, entity_rows, row_count)
        return False
    if field_data.type in (
        DataType.FLOAT_VECTOR,
        DataType.FLOAT16_VECTOR,
        DataType.BFLOAT16_VECTOR,
        DataType.BINARY_VECTOR,
        DataType.SPARSE_FLOAT_VECTOR,
        DataType.INT8_VECTOR,
    ):
        return True
    if field_data.type == DataType._ARRAY_OF_STRUCT:
        return True
    if field_data.type == DataType._ARRAY_OF_VECTOR:
        return True
    if field_data.type == DataType.STRING:
        raise MilvusException(message="Not support string yet")
    return False


def extract_vector_array_row_data(field_data: Any, index: int):
    array = field_data.vectors.vector_array.data[index]
    element_type = field_data.vectors.vector_array.element_type

    if element_type == DataType.FLOAT_VECTOR:
        return list(np.array(array.float_vector.data, dtype=np.float32))

    if element_type == DataType.FLOAT16_VECTOR:
        byte_data = array.float16_vector
        return list(np.frombuffer(byte_data, dtype=np.float16))

    if element_type == DataType.BFLOAT16_VECTOR:
        byte_data = array.bfloat16_vector
        return list(
            np.frombuffer(byte_data, dtype="bfloat16" if hasattr(np, "bfloat16") else np.uint16)
        )

    if element_type == DataType.INT8_VECTOR:
        byte_data = array.int8_vector
        return list(np.frombuffer(byte_data, dtype=np.int8))

    if element_type == DataType.BINARY_VECTOR:
        return [array.binary_vector]

    raise ParamError(message=f"Unimplemented type: {element_type} for vector array extraction")


# pylint: disable=R1702 (too-many-nested-blocks)
# pylint: disable=R0915 (too-many-statements)
def extract_row_data_from_fields_data(
    fields_data: Any,
    index: Any,
    dynamic_output_fields: Optional[List] = None,
):
    if not fields_data:
        return {}

    entity_row_data = {}
    dynamic_fields = dynamic_output_fields or set()

    def check_append(field_data: Any, row_data: Dict):
        if field_data.type == DataType.STRING:
            raise MilvusException(message="Not support string yet")

        if field_data.type == DataType.BOOL and len(field_data.scalars.bool_data.data) >= index:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = field_data.scalars.bool_data.data[index]
            return

        if (
            field_data.type in (DataType.INT8, DataType.INT16, DataType.INT32)
            and len(field_data.scalars.int_data.data) >= index
        ):
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = field_data.scalars.int_data.data[index]
            return

        if field_data.type == DataType.INT64 and len(field_data.scalars.long_data.data) >= index:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = field_data.scalars.long_data.data[index]
            return

        if field_data.type == DataType.FLOAT and len(field_data.scalars.float_data.data) >= index:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = np.single(field_data.scalars.float_data.data[index])
            return

        if field_data.type == DataType.DOUBLE and len(field_data.scalars.double_data.data) >= index:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = field_data.scalars.double_data.data[index]
            return

        if (
            field_data.type == DataType.VARCHAR
            and len(field_data.scalars.string_data.data) >= index
        ):
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = field_data.scalars.string_data.data[index]
            return

        if (
            field_data.type == DataType.GEOMETRY
            and len(field_data.scalars.geometry_wkt_data.data) >= index
        ):
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                entity_row_data[field_data.field_name] = None
                return
            entity_row_data[field_data.field_name] = field_data.scalars.geometry_wkt_data.data[
                index
            ]
            return

        if field_data.type == DataType.JSON and len(field_data.scalars.json_data.data) >= index:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            try:
                json_dict = orjson.loads(field_data.scalars.json_data.data[index])
            except Exception as e:
                logger.error(
                    f"extract_row_data_from_fields_data::Failed to load JSON data: {e}, original data: {field_data.scalars.json_data.data[index]}"
                )
                raise

            if not field_data.is_dynamic:
                row_data[field_data.field_name] = json_dict
                return

            if not dynamic_fields:
                row_data.update(json_dict)
                return

            row_data.update({k: v for k, v in json_dict.items() if k in dynamic_fields})
            return
        if field_data.type == DataType.ARRAY and len(field_data.scalars.array_data.data) >= index:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            row_data[field_data.field_name] = extract_array_row_data(field_data, index)

        elif field_data.type == DataType.FLOAT_VECTOR:
            dim = field_data.vectors.dim
            if len(field_data.vectors.float_vector.data) >= index * dim:
                start_pos, end_pos = index * dim, (index + 1) * dim
                # Here we use numpy.array to convert the float64 values to numpy.float32 values,
                # and return a list of numpy.float32 to users
                # By using numpy.array, performance improved by 60% for topk=16384 dim=1536 case.
                arr = np.array(
                    field_data.vectors.float_vector.data[start_pos:end_pos], dtype=np.float32
                )
                row_data[field_data.field_name] = list(arr)
        elif field_data.type == DataType.BINARY_VECTOR:
            dim = field_data.vectors.dim
            if len(field_data.vectors.binary_vector) >= index * (dim // 8):
                start_pos, end_pos = index * (dim // 8), (index + 1) * (dim // 8)
                row_data[field_data.field_name] = [
                    field_data.vectors.binary_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType.BFLOAT16_VECTOR:
            dim = field_data.vectors.dim
            if len(field_data.vectors.bfloat16_vector) >= index * (dim * 2):
                start_pos, end_pos = index * (dim * 2), (index + 1) * (dim * 2)
                row_data[field_data.field_name] = [
                    field_data.vectors.bfloat16_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType.FLOAT16_VECTOR:
            dim = field_data.vectors.dim
            if len(field_data.vectors.float16_vector) >= index * (dim * 2):
                start_pos, end_pos = index * (dim * 2), (index + 1) * (dim * 2)
                row_data[field_data.field_name] = [
                    field_data.vectors.float16_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType.SPARSE_FLOAT_VECTOR:
            row_data[field_data.field_name] = sparse_parse_single_row(
                field_data.vectors.sparse_float_vector.contents[index]
            )
        elif field_data.type == DataType.INT8_VECTOR:
            dim = field_data.vectors.dim
            if len(field_data.vectors.int8_vector) >= index * dim:
                start_pos, end_pos = index * dim, (index + 1) * dim
                row_data[field_data.field_name] = [
                    field_data.vectors.int8_vector[start_pos:end_pos]
                ]
        elif (
            field_data.type == DataType._ARRAY_OF_VECTOR
            and len(field_data.vectors.vector_array.data) >= index
        ):
            row_data[field_data.field_name] = extract_vector_array_row_data(field_data, index)
        elif field_data.type == DataType._ARRAY_OF_STRUCT:
            row_data[field_data.field_name] = {}
            for sub_field_data in field_data.struct_arrays.fields:
                check_append(sub_field_data, row_data[field_data.field_name])

    for field_data in fields_data:
        check_append(field_data, entity_row_data)

    return entity_row_data


def get_array_length(array_item: Any) -> int:
    """Get the length of an array field from its data."""
    if hasattr(array_item, "string_data") and array_item.string_data:
        return len(array_item.string_data.data)
    if hasattr(array_item, "int_data") and array_item.int_data:
        return len(array_item.int_data.data)
    if hasattr(array_item, "long_data") and array_item.long_data:
        return len(array_item.long_data.data)
    if hasattr(array_item, "float_data") and array_item.float_data:
        return len(array_item.float_data.data)
    if hasattr(array_item, "double_data") and array_item.double_data:
        return len(array_item.double_data.data)
    if hasattr(array_item, "bool_data") and array_item.bool_data:
        return len(array_item.bool_data.data)
    return 0


def get_array_value_at_index(array_item: Any, idx: int) -> Any:
    """Get the value at a specific index from an array field."""
    if (
        hasattr(array_item, "string_data")
        and array_item.string_data
        and idx < len(array_item.string_data.data)
    ):
        return array_item.string_data.data[idx]
    if (
        hasattr(array_item, "int_data")
        and array_item.int_data
        and idx < len(array_item.int_data.data)
    ):
        return array_item.int_data.data[idx]
    if (
        hasattr(array_item, "long_data")
        and array_item.long_data
        and idx < len(array_item.long_data.data)
    ):
        return array_item.long_data.data[idx]
    if (
        hasattr(array_item, "float_data")
        and array_item.float_data
        and idx < len(array_item.float_data.data)
    ):
        return array_item.float_data.data[idx]
    if (
        hasattr(array_item, "double_data")
        and array_item.double_data
        and idx < len(array_item.double_data.data)
    ):
        return array_item.double_data.data[idx]
    if (
        hasattr(array_item, "bool_data")
        and array_item.bool_data
        and idx < len(array_item.bool_data.data)
    ):
        return array_item.bool_data.data[idx]
    return None


def extract_struct_array_from_column_data(struct_arrays: Any, row_idx: int) -> List[Dict[str, Any]]:
    """Convert column-format struct data back to array of structs format.

    Milvus stores struct arrays in column format where each field's data is stored separately.
    This function converts it back to row format for user consumption.

    For example, if the original one row of array of struct data was:
    [
        {"name": "Alice", "age": 30, "vector": [1.0, 2.0]},
        {"name": "Bob", "age": 25, "vector": [3.0, 4.0]}
    ]

    Milvus stores it as:
    - name field: ["Alice", "Bob"]
    - age field: [30, 25]
    - vector field: [[1.0, 2.0], [3.0, 4.0]]

    This function reconstructs the original array of struct format.

    Args:
        struct_arrays: The struct_arrays field containing column-format data with sub-fields
        row_idx: The row index to extract data for

    Returns:
        List of dictionaries representing the struct array in row format
    """
    if not struct_arrays or not hasattr(struct_arrays, "fields"):
        return []

    # Determine the number of struct elements by checking the first sub-field's data length
    # All sub-fields should have the same number of elements
    num_structs = 0
    for sub_field in struct_arrays.fields:
        if sub_field.type == DataType.ARRAY and row_idx < len(sub_field.scalars.array_data.data):
            array_item = sub_field.scalars.array_data.data[row_idx]
            num_structs = get_array_length(array_item)
            if num_structs > 0:
                break
        elif sub_field.type == DataType._ARRAY_OF_VECTOR:
            if (
                hasattr(sub_field, "vectors")
                and hasattr(sub_field.vectors, "vector_array")
                and row_idx < len(sub_field.vectors.vector_array.data)
            ):
                vector_data = sub_field.vectors.vector_array.data[row_idx]
                dim = vector_data.dim
                element_type = sub_field.vectors.vector_array.element_type

                if element_type == DataType.FLOAT_VECTOR:
                    num_structs = len(vector_data.float_vector.data) // dim
                elif element_type == DataType.FLOAT16_VECTOR:
                    num_structs = len(vector_data.float16_vector) // (dim * 2)
                elif element_type == DataType.BFLOAT16_VECTOR:
                    num_structs = len(vector_data.bfloat16_vector) // (dim * 2)
                elif element_type == DataType.INT8_VECTOR:
                    num_structs = len(vector_data.int8_vector) // dim
                elif element_type == DataType.BINARY_VECTOR:
                    num_structs = len(vector_data.binary_vector) // (dim // 8)
                else:
                    num_structs = 0

                if num_structs > 0:
                    break

    # Build array of struct objects by extracting data at each struct index
    struct_array = []
    for struct_idx in range(num_structs):
        struct_obj = {}

        for sub_field in struct_arrays.fields:
            sub_field_name = sub_field.field_name

            # Handle scalar array fields (VARCHAR, INT, FLOAT, etc.)
            if sub_field.type == DataType.ARRAY:
                if row_idx < len(sub_field.scalars.array_data.data):
                    array_item = sub_field.scalars.array_data.data[row_idx]
                    # Extract the value at struct_idx from the appropriate data type
                    struct_obj[sub_field_name] = get_array_value_at_index(array_item, struct_idx)

            elif sub_field.type == DataType._ARRAY_OF_VECTOR:
                if hasattr(sub_field, "vectors") and hasattr(sub_field.vectors, "vector_array"):
                    vector_array = sub_field.vectors.vector_array
                    if row_idx < len(vector_array.data):
                        vector_data = vector_array.data[row_idx]
                        dim = vector_data.dim
                        element_type = vector_array.element_type

                        if element_type == DataType.FLOAT_VECTOR:
                            float_data = vector_data.float_vector.data
                            vec_start = struct_idx * dim
                            vec_end = vec_start + dim
                            if vec_end <= len(float_data):
                                struct_obj[sub_field_name] = list(float_data[vec_start:vec_end])
                            else:
                                struct_obj[sub_field_name] = None

                        elif element_type == DataType.FLOAT16_VECTOR:
                            byte_data = vector_data.float16_vector
                            bytes_per_vec = dim * 2
                            vec_start = struct_idx * bytes_per_vec
                            vec_end = vec_start + bytes_per_vec
                            if vec_end <= len(byte_data):
                                vec_bytes = byte_data[vec_start:vec_end]
                                struct_obj[sub_field_name] = list(
                                    np.frombuffer(vec_bytes, dtype=np.float16)
                                )
                            else:
                                struct_obj[sub_field_name] = None

                        elif element_type == DataType.BFLOAT16_VECTOR:
                            byte_data = vector_data.bfloat16_vector
                            bytes_per_vec = dim * 2
                            vec_start = struct_idx * bytes_per_vec
                            vec_end = vec_start + bytes_per_vec
                            if vec_end <= len(byte_data):
                                vec_bytes = byte_data[vec_start:vec_end]
                                dtype = "bfloat16" if hasattr(np, "bfloat16") else np.uint16
                                struct_obj[sub_field_name] = list(
                                    np.frombuffer(vec_bytes, dtype=dtype)
                                )
                            else:
                                struct_obj[sub_field_name] = None

                        elif element_type == DataType.INT8_VECTOR:
                            byte_data = vector_data.int8_vector
                            bytes_per_vec = dim
                            vec_start = struct_idx * bytes_per_vec
                            vec_end = vec_start + bytes_per_vec
                            if vec_end <= len(byte_data):
                                vec_bytes = byte_data[vec_start:vec_end]
                                struct_obj[sub_field_name] = list(
                                    np.frombuffer(vec_bytes, dtype=np.int8)
                                )
                            else:
                                struct_obj[sub_field_name] = None

                        elif element_type == DataType.BINARY_VECTOR:
                            byte_data = vector_data.binary_vector
                            bytes_per_vec = dim // 8
                            vec_start = struct_idx * bytes_per_vec
                            vec_end = vec_start + bytes_per_vec
                            if vec_end <= len(byte_data):
                                struct_obj[sub_field_name] = [byte_data[vec_start:vec_end]]
                            else:
                                struct_obj[sub_field_name] = None
                        else:
                            # Unsupported vector type, set to None
                            struct_obj[sub_field_name] = None

            # All struct sub-fields should be either ARRAY or ARRAY_OF_VECTOR
            # Any other type indicates a bug in the data conversion logic
            else:
                raise ParamError(
                    message=f"Unexpected field type {sub_field.type} for struct sub-field {sub_field_name}. "
                    f"Struct fields must be ARRAY or ARRAY_OF_VECTOR internally."
                )

        struct_array.append(struct_obj)

    return struct_array
